import base64

import boto3
from mypy_boto3_kms.client import KMSClient


def get_kms_client() -> KMSClient:
    """
    Returns a KMS client
    """
    return boto3.client("kms")


def encrypt(key_id: str, value: str, kms: KMSClient) -> str:
    """
    Takes a KMS key id, encrypts the value using it and
    returns the encrypted value
    """
    response = kms.encrypt(KeyId=key_id, Plaintext=value.encode())
    return base64.b64encode(response["CiphertextBlob"]).decode()


def decrypt(key_id: str, value: str, kms: KMSClient) -> str:
    """
    Takes a KMS key id, decrypts the value using it and
    returns the decrypted value
    """
    response = kms.decrypt(
        KeyId=key_id,
        CiphertextBlob=bytes(base64.b64decode(value)),
    )
    return response["Plaintext"].decode()



import json
from typing import Any
from typing import Callable
from typing import List
from typing import Optional
from typing import Type

import boto3
import yaml
from mypy_boto3_appconfigdata.client import AppConfigDataClient
from mypy_boto3_kms.client import KMSClient
from sitri import Sitri
from sitri.providers.base import ConfigProvider
from sitri.providers.base import PathModeStateProvider
from sitri.providers.contrib.system import SystemConfigProvider
from sitri.providers.contrib.yaml import YamlConfigProvider
from sitri.strategy.index_priority import IndexPriorityStrategy


system = SystemConfigProvider(prefix="TWO")

yaml = YamlConfigProvider(yaml_path="./data.yaml", default_separator="/")
strategy = IndexPriorityStrategy([system, yaml])


class AwsAppConfigProvider(PathModeStateProvider, ConfigProvider):
    """Config provider for AWS AppConfig storage."""

    provider_code = "app_config"
    path_seperator = "."
    client: AppConfigDataClient

    def __init__(
        self,
        application: str,
        environment: str,
        profile: str,
        update_interval: int = 300,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        :param vedis_connector: return connection to vedis
        :param hash_name: name for hash (key-value object) in vedis
        """
        super().__init__(*args, **kwargs)

        self.application = application
        self.environment = environment
        self.profile = profile
        self.update_interval = update_interval

        self._poll_interval = None
        self._next_config_token = None

        self._time_last_updated = None
        self._boto3_app_config_client_instance = None
        self._config = None
        self._raw_config = None
        self._content_type = None

        self._init_configuration()

    @property
    def client(self) -> AppConfigDataClient:
        if self._boto3_app_config_client_instance is None:
            self._boto3_app_config_client_instance = boto3.client("appconfigdata")
        return self._boto3_app_config_client_instance

    def _init_configuration(self):
        resp = self.client.start_configuration_session(
            ApplicationIdentifier=self.application,
            ConfigurationProfileIdentifier=self.profile,
            EnvironmentIdentifier=self.environment,
            RequiredMinimumPollIntervalInSeconds=self.update_interval,
        )
        self._next_config_token = resp["InitialConfigurationToken"]

    def _update_configuration(self):
        response = self.client.get_latest_configuration(
            ConfigurationToken=self._next_config_token,
        )
        self._next_config_token = response["NextPollConfigurationToken"]
        self._poll_interval = int(response["NextPollIntervalInSeconds"])
        self._time_last_updated = response["LastUpdatedTime"]
        content = response["Configuration"].read()  # type: bytes
        if content == b"":
            return False

        if response["ContentType"] == "application/x-yaml":
            try:
                self._config = yaml.safe_load(content)
            except yaml.YAMLError as error:
                message = "Unable to parse YAML configuration data"
                if hasattr(error, "problem_mark"):
                    message = (
                        f"{message} at line {error.problem_mark.line + 1} "
                        f"column {error.problem_mark.column + 1}"
                    )
                raise ValueError(message) from error
        elif response["ContentType"] == "application/json":
            try:
                self._config = json.loads(content.decode("utf-8"))
            except json.JSONDecodeError as error:
                raise ValueError(error.msg) from error
        elif response["ContentType"] == "text/plain":
            self._config = content.decode("utf-8")
        else:
            self._config = content

        self._raw_config = content
        self._content_type = response["ContentType"]
        return True

    def get(
        self,
        key: str,
        path_mode: Optional[bool] = None,
        separator: Optional[str] = None,
        **kwargs,
    ) -> Optional[Any]:
        """Get value from config.

        :param key: key or path for search
        :param path_mode: boolean mode switcher
        :param separator: separator for path keys in path mode
        """
        separator = separator if separator else self.path_seperator

        if self._get_path_mode_state(path_mode):
            return self._get_by_path(key, separator=separator)

        try:
            value = self._config[key]
        except KeyError:
            value = None
        return value

    def _get_by_path(self, path: str, separator: str) -> Any:
        """Retrieve value from a dictionary using a list of keys.

        :param path: string with separated keys
        """
        dict_local = self._config.copy()
        keys = path.split(separator)

        for key in keys:
            try:
                dict_local = dict_local[int(key)] if key.isdigit() else dict_local[key]
            except Exception:
                if key not in dict_local.keys():
                    return None

                dict_local = dict_local[key]
        return dict_local

    def keys(
        self,
        path_mode: bool = False,
        separator: Optional[str] = None,
        **kwargs: Any,
    ) -> List[str]:
        """Keys in config.

        :param path_mode: [future] path mode for keys list
        :param separator: [future] separators for keys in path mode
        """
        if not path_mode:
            return list(self._config.keys())
        raise NotImplementedError("Path-mode not implemented!")


# ENV = system.get("env")

# conf = Sitri(
#     config_provider=,
# )

# print(system.get_config("host"))
# # Output: example.com

# print(system.config.keys())
# # Output: ["host", "password"]


# yaml.get("test.test_key1", ":(")
# # Output: :(

# yaml.get("test.test_key1", ":(", path_mode=True)
# # Output: :(

# yaml.get("test.test_key1", ":(", path_mode=True, separator=".")
# # Output: 1

# yaml.get("test/test_key1", ":(", path_mode=True)
# # Output: 1

# yaml.get("test0")
# # Output: 0



from __future__ import annotations

from typing import List
from typing import Optional

from sqlalchemy import Column
from sqlalchemy import create_engine
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import select
from sqlalchemy import String
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import registry
from sqlalchemy.orm import relationship
from sqlalchemy.orm import Session


engine = create_engine("sqlite://", echo=True)


class Base(DeclarativeBase):
    pass


class User(Base):
    __tablename__ = "user_account"

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(String(30))
    fullname: Mapped[Optional[str]]

    addresses: Mapped[list["Address"]] = relationship(
        back_populates="user", cascade="all, delete-orphan"
    )

    def __repr__(self) -> str:
        return f"User(id={self.id!r}, name={self.name!r}, fullname={self.fullname!r})"


class Address(Base):
    __tablename__ = "address"

    id: Mapped[int] = mapped_column(primary_key=True)
    email_address: Mapped[str]
    user_id: Mapped[int] = mapped_column(ForeignKey("user_account.id"))

    user: Mapped["User"] = relationship(back_populates="addresses")

    def __repr__(self) -> str:
        return f"Address(id={self.id!r}, email_address={self.email_address!r})"


with Session(engine) as session:
    spongebob = User(
        name="spongebob",
        fullname="Spongebob Squarepants",
        addresses=[Address(email_address="spongebob@sqlalchemy.org")],
    )

    session.add(spongebob)
    session.commit()

session = Session(engine)

stmt = select(User).where(User.name.in_(["spongebob"]))

for user in session.scalars(stmt):
    print(user)



from pallets_sphinx_themes import get_version
from pallets_sphinx_themes import ProjectLink

# Project --------------------------------------------------------------

project = "Flask-SQLAlchemy"
copyright = "2010 Pallets"
author = "Pallets"
release, version = get_version("Flask-SQLAlchemy")

# General --------------------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "pallets_sphinx_themes",
    "sphinx_issues",
]
autodoc_typehints = "description"
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "flask": ("https://flask.palletsprojects.com/", None),
    "sqlalchemy": ("https://docs.sqlalchemy.org/", None),
}
issues_github_path = "pallets-eco/flask-sqlalchemy"

# HTML -----------------------------------------------------------------

html_theme = "flask"
html_context = {
    "project_links": [
        ProjectLink("Donate", "https://palletsprojects.com/donate"),
        ProjectLink("PyPI Releases", "https://pypi.org/project/Flask-SQLAlchemy/"),
        ProjectLink("Source Code", "https://github.com/pallets-eco/flask-sqlalchemy/"),
        ProjectLink(
            "Issue Tracker", "https://github.com/pallets-eco/flask-sqlalchemy/issues/"
        ),
        ProjectLink("Website", "https://palletsprojects.com/"),
        ProjectLink("Twitter", "https://twitter.com/PalletsTeam"),
        ProjectLink("Chat", "https://discord.gg/pallets"),
    ]
}
html_sidebars = {
    "index": ["project.html", "localtoc.html", "searchbox.html"],
    "**": ["localtoc.html", "relations.html", "searchbox.html"],
}
singlehtml_sidebars = {"index": ["project.html", "localtoc.html"]}
html_static_path = ["_static"]
html_favicon = "_static/flask-sqlalchemy-logo.png"
html_logo = "_static/flask-sqlalchemy-logo.png"
html_title = f"Flask-SQLAlchemy Documentation ({version})"
html_show_sourcelink = False



import json
import os
from typing import Callable
from typing import Dict
from typing import Generic
from typing import Sequence
from typing import Tuple
from typing import TypeVar
from typing import Union


Exists = bool
JsonSerializable = Union[Dict, Sequence, str, bool, int, float]
JsonSerializableGeneric = TypeVar("JsonSerializableGeneric", bound=JsonSerializable)


class DictJsonRepo(Generic[JsonSerializableGeneric]):
    def __init__(self, directory: str = None) -> None:
        self.directory = directory or os.getcwd()

    def read(self, filename: str) -> JsonSerializableGeneric:
        with open(self._build_path(filename), encoding="utf-8") as f:
            return json.load(f)

    def write(self, data: JsonSerializableGeneric, filename: str) -> JsonSerializableGeneric:
        with open(self._build_path(filename), "w", encoding="utf-8") as f:
            json.dump(data, f)

        return data

    def read_or_write_default(
        self, filename: str, default_factory: Callable[[], JsonSerializableGeneric]
    ) -> Tuple[JsonSerializableGeneric, Exists]:
        if os.path.exists(self._build_path(filename)):
            return self.read(filename), True
        else:
            return self.write(default_factory(), filename), False

    def _build_path(self, filename: str) -> str:
        return os.path.join(self.directory, filename)



import functools

from quart import Blueprint
from quart import flash
from quart import g
from quart import redirect
from quart import render_template
from quart import request
from quart import session
from quart import url_for

from quartr import db
from quartr.auth.models import User

bp = Blueprint("auth", __name__, url_prefix="/auth")


def login_required(view):
    """View decorator that redirects anonymous users to the login page."""

    @functools.wraps(view)
    def wrapped_view(**kwargs):
        if g.user is None:
            return redirect(url_for("auth.login"))

        return view(**kwargs)

    return wrapped_view


@bp.before_app_request
def load_logged_in_user():
    """If a user id is stored in the session, load the user object from
    the database into ``g.user``."""
    user_id = session.get("user_id")

    if user_id is not None:
        g.user = db.session.get(User, user_id)
    else:
        g.user = None


@bp.route("/register", methods=("GET", "POST"))
def register():
    """Register a new user.

    Validates that the username is not already taken. Hashes the
    password for security.
    """
    if request.method == "POST":
        username = request.form["username"]
        password = request.form["password"]
        error = None

        if not username:
            error = "Username is required."
        elif not password:
            error = "Password is required."
        elif db.session.execute(
            db.select(db.select(User).filter_by(username=username).exists())
        ).scalar():
            error = f"User {username} is already registered."

        if error is None:
            # the name is available, create the user and go to the login page
            db.session.add(User(username=username, password=password))
            db.session.commit()
            return redirect(url_for("auth.login"))

        flash(error)

    return render_template("auth/register.html")


@bp.route("/login", methods=("GET", "POST"))
def login():
    """Log in a registered user by adding the user id to the session."""
    if request.method == "POST":
        username = request.form["username"]
        password = request.form["password"]
        error = None
        select = db.select(User).filter_by(username=username)
        user = db.session.execute(select).scalar()

        if user is None:
            error = "Incorrect username."
        elif not user.check_password(password):
            error = "Incorrect password."

        if error is None:
            # store the user id in a new session and return to the index
            session.clear()
            session["user_id"] = user.id
            return redirect(url_for("index"))

        flash(error)

    return render_template("auth/login.html")


@bp.route("/logout")
def logout():
    """Clear the current session, including the stored user id."""
    session.clear()
    return redirect(url_for("index"))



import pytest

from quartr import db
from quartr.auth.models import User
from quartr.blog.models import Post


def test_index(client, auth):
    text = client.get("/").text
    assert "Log In" in text
    assert "Register" in text

    auth.login()
    text = client.get("/").text
    assert "test title" in text
    assert "by test on 2018-01-01" in text
    assert "test\nbody" in text
    assert 'href="/1/update"' in text


@pytest.mark.parametrize("path", ("/create", "/1/update", "/1/delete"))
def test_login_required(client, path):
    response = client.post(path)
    assert response.headers["Location"] == "/auth/login"


def test_author_required(app, client, auth):
    # change the post author to another user
    with app.app_context():
        db.session.get(Post, 1).author = db.session.get(User, 2)
        db.session.commit()

    auth.login()
    # current user can't modify other user's post
    assert client.post("/1/update").status_code == 403
    assert client.post("/1/delete").status_code == 403
    # current user doesn't see edit link
    assert 'href="/1/update"' not in client.get("/").text


@pytest.mark.parametrize("path", ("/2/update", "/2/delete"))
def test_exists_required(client, auth, path):
    auth.login()
    assert client.post(path).status_code == 404


def test_create(client, auth, app):
    auth.login()
    assert client.get("/create").status_code == 200
    client.post("/create", data={"title": "created", "body": ""})

    with app.app_context():
        select = db.select(db.func.count(Post.id))
        post_count = db.session.execute(select).scalar()
        assert post_count == 2


def test_update(client, auth, app):
    auth.login()
    assert client.get("/1/update").status_code == 200
    client.post("/1/update", data={"title": "updated", "body": ""})

    with app.app_context():
        assert db.session.get(Post, 1).title == "updated"


@pytest.mark.parametrize("path", ("/create", "/1/update"))
def test_create_update_validate(client, auth, path):
    auth.login()
    response = client.post(path, data={"title": "", "body": ""})
    assert "Title is required." in response.text


def test_delete(client, auth, app):
    auth.login()
    response = client.post("/1/delete")
    assert response.headers["Location"] == "/"

    with app.app_context():
        assert db.session.get(Post, 1) is None



from __future__ import annotations

import pytest
import sqlalchemy as sa
from quart import Quart

from quart_sqlalchemy import SQLAlchemy
from quart_sqlalchemy.session import Session


async def test_scope(app: Quart, db: SQLAlchemy) -> None:
    with pytest.raises(RuntimeError):
        db.session()

    async with app.app_context():
        first = db.session()
        second = db.session()
        assert first is second
        assert isinstance(first, Session)

    async with app.app_context():
        third = db.session()
        assert first is not third


async def test_custom_scope(app: Quart) -> None:
    count = 0

    def scope() -> int:
        nonlocal count
        count += 1
        return count

    db = SQLAlchemy(app, session_options={"scopefunc": scope})

    async with app.app_context():
        first = db.session()
        second = db.session()
        assert first is not second  # a new scope is generated on each call
        first.close()
        second.close()


async def test_session_class(app: Quart) -> None:
    class CustomSession(Session):
        pass
    
    async with app.app_context():
        db = SQLAlchemy(app, session_options={"class_": CustomSession})
        assert isinstance(db.session(), CustomSession)


async def test_session_uses_bind_key(app: Quart) -> None:
    async with app.app_context():
        app.config["SQLALCHEMY_BINDS"] = {"a": "sqlite://"}
        db = SQLAlchemy(app)

        class User(db.Model):
            id = sa.Column(sa.Integer, primary_key=True)

        class Post(db.Model):
            __bind_key__ = "a"
            id = sa.Column(sa.Integer, primary_key=True)

        assert db.session.get_bind(mapper=User) is db.engine
        assert db.session.get_bind(mapper=Post) is db.engines["a"]



from __future__ import annotations

import asyncio
import typing as t
from pathlib import Path

import pytest
import sqlalchemy as sa
from quart import Quart

from quart_sqlalchemy import SQLAlchemy


@pytest.fixture(scope="module")
def event_loop():
    policy = asyncio.get_event_loop_policy()
    loop = policy.new_event_loop()
    yield loop
    loop.close()


@pytest.fixture
def app(request: pytest.FixtureRequest, tmp_path: Path) -> Quart:
    app = Quart(request.module.__name__, instance_path=str(tmp_path / "instance"))
    app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite://"
    app.config["SQLALCHEMY_RECORD_QUERIES"] = False
    return app


@pytest.fixture
async def db(app: Quart) -> SQLAlchemy:
    yield SQLAlchemy(app)


@pytest.fixture
async def Todo(app: Quart, db: SQLAlchemy) -> t.Any:
    class Todo(db.Model):
        id = sa.Column(sa.Integer, primary_key=True)
        title = sa.Column(sa.String)

    async with app.app_context():
        db.create_all()

    yield Todo

    async with app.app_context():
        db.drop_all()



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from enum import Enum
from inspect import getmodule
from types import UnionType
from typing import Annotated
from typing import Any
from typing import Callable
from typing import cast
from typing import ClassVar
from typing import Generic
from typing import get_args
from typing import get_origin
from typing import get_type_hints
from typing import Iterable
from typing import TYPE_CHECKING
from typing import TypeVar
from typing import Union

from pydantic import BaseModel
from pydantic import create_model
from pydantic import validator
from pydantic.fields import FieldInfo
from sqlalchemy import inspect
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import Mapped


if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Iterable
    from typing import Any
    from typing import Literal

    from pydantic.fields import FieldInfo
    from pydantic.typing import AnyClassMethod
    from sqlalchemy import Column
    from sqlalchemy.orm import Mapper
    from sqlalchemy.orm import RelationshipProperty
    from sqlalchemy.sql.base import ReadOnlyColumnCollection
    from sqlalchemy.util import ReadOnlyProperties


AnyDeclarative = TypeVar("AnyDeclarative", bound=DeclarativeBase)


class Mark(str, Enum):
    """For marking column definitions on the domain models.

    Example:
    ```python
    class Model(Base):
        ...
        updated_at: Mapped[datetime] = mapped_column(info={"dto": Mark.READ_ONLY})
    ```
    """

    READ_ONLY = "read-only"
    """To mark a field that can be read, but not updated by clients."""
    PRIVATE = "private"
    """To mark a field that can neither be read or updated by clients."""


class Purpose(str, Enum):
    """For identifying the purpose of a DTO to the factory.

    The factory will exclude fields marked as private or read-only on the domain model depending
    on the purpose of the DTO.

    Example:
    ```python
    ReadDTO = dto.factory("AuthorReadDTO", Author, purpose=dto.Purpose.READ)
    ```
    """

    READ = "read"
    """To mark a DTO that is to be used to serialize data returned to clients."""
    WRITE = "write"
    """To mark a DTO that is to deserialize and validate data provided by clients."""


@dataclass
class DTOField:
    """For configuring DTO behavior on SQLAlchemy model fields."""

    mark: Mark | None = None
    """Mark the field as read-only, or private."""
    pydantic_type: Any | None = None
    """Override the field type on the pydantic model for this attribute."""
    pydantic_field: FieldInfo | None = None
    """If provided, used for the pydantic model for this attribute."""
    validators: Iterable[Callable[[Any], Any]] | None = None
    """Single argument callables that are defined on the DTO as validators for the field."""


@dataclass
class DTOConfig:
    """Control the generated DTO."""

    purpose: Purpose
    """Configure the DTO for "read" or "write" operations."""
    exclude: set[str] = field(default_factory=set)
    """Explicitly exclude fields from the generated DTO."""


def config(
    purpose: Purpose | Literal["read", "write"], exclude: set[str] | None = None
) -> DTOConfig:
    """
    Args:
        purpose: Is the DTO for parsing "write" data, or serializing "read" data?
        exclude: Omit fields from dto by key name.

    Returns:
        `DTOConfig` object configured per parameters.
    """
    exclude = set() if exclude is None else exclude
    return DTOConfig(purpose=Purpose(purpose), exclude=exclude)


def field(
    mark: Mark | Literal["read-only", "private"] | None = None,
    pydantic_type: Any | None = None,
    pydantic_field: FieldInfo | None = None,
    validators: Iterable[Callable[[Any], Any]] | None = None,
) -> dict[str, DTOField]:
    """Create `dto.DTOField()` wrapped in a dict for SQLAlchemy info field.

    Args:
        mark: How this field should be treated by the model factory.
        pydantic_type: Override the type annotation for this field.
        pydantic_field: Result of Pydantic's `DTOField()` function. Override the `FieldInfo` instance
            used by the generated model.
        validators: Added to the generated model as validators, with `allow_reuse=True`.
    """
    return {
        "dto": DTOField(
            mark=Mark(mark) if mark is not None else mark,
            pydantic_type=pydantic_type,
            pydantic_field=pydantic_field,
            validators=validators,
        )
    }


class FromMapped(BaseModel, Generic[AnyDeclarative]):
    """Produce an SQLAlchemy instance with values from a pydantic model."""

    __sqla_model__: ClassVar[type[DeclarativeBase]]

    class Config:
        """Set orm_mode for `to_mapped()` method."""

        orm_mode = True

    def __class_getitem__(
        cls, item: Annotated[type[AnyDeclarative], DTOConfig | Literal["read", "write"]]
    ) -> type[FromMapped[AnyDeclarative]]:
        """Decorate `cls` with result from `factory()`.

        Args:
            item: Can be either of a SQLAlchemy ORM instance, or a `typing.Annotated` annotation
                where the first argument is a SQLAlchemy ORM instance, and the second is an instance
                of `DTOConfig`.

        Returns:
            A new Pydantic model type, with `cls` as its base class, and additional fields derived
            from the SQLAlchemy model, respecting any declared configuration.
        """
        if get_origin(item) is Annotated:
            model, pos_arg, *_ = get_args(item)
            if isinstance(pos_arg, str):
                dto_config = config(pos_arg)  # type:ignore[arg-type]
            else:
                dto_config = pos_arg
        else:
            raise ValueError("Unexpected type annotation for `FromMapped`.")
        return cls._factory(
            cls.__name__,
            cast("type[AnyDeclarative]", model),
            dto_config.purpose,
            exclude=dto_config.exclude,
        )

    # pylint: disable=arguments-differ
    def __init_subclass__(cls, model: type[AnyDeclarative] | None = None, **kwargs: Any) -> None:
        """Set `__sqla_model__` on type.

        Args:
            model: Model represented by the DTO
            kwargs: Passed to `super().__init_subclass__()`
        """
        super().__init_subclass__(**kwargs)
        if model is not None:
            cls.__sqla_model__ = model

    def to_mapped(self) -> AnyDeclarative:
        """Create an instance of `self.__sqla_model__`

        Fill the bound SQLAlchemy model recursively with values from
        this dataclass.
        """
        as_model = {}
        for pydantic_field in self.__fields__.values():
            value = getattr(self, pydantic_field.name)
            if isinstance(value, (list, tuple)):
                value = [el.to_mapped() if isinstance(el, FromMapped) else el for el in value]
            if isinstance(value, FromMapped):
                value = value.to_mapped()
            as_model[pydantic_field.name] = value
        return cast("AnyDeclarative", self.__sqla_model__(**as_model))

    @classmethod
    def _factory(
        cls, name: str, model: type[DeclarativeBase], purpose: Purpose, exclude: set[str]
    ) -> type[FromMapped[AnyDeclarative]]:
        exclude = set() if exclude is None else exclude

        columns, relationships = _inspect_model(model)
        fields: dict[str, tuple[Any, FieldInfo]] = {}
        validators: dict[str, AnyClassMethod] = {}
        for key, type_hint in get_type_hints(model, localns=_get_localns(model)).items():
            if get_origin(type_hint) is Mapped:
                (type_hint,) = get_args(type_hint)

            elem: Column | RelationshipProperty
            if key in columns:
                elem = columns[key]
            elif key in relationships:
                elem = relationships[key]
            else:
                # class var, anything else??
                continue

            dto_field = _get_dto_field(elem)

            if _should_exclude_field(purpose, elem, exclude, dto_field):
                continue

            if dto_field.pydantic_type is not None:
                type_hint = dto_field.pydantic_type

            for i, func in enumerate(dto_field.validators or []):
                validators[f"_validates_{key}_{i}"] = validator(key, allow_reuse=True)(func)

            type_hint = cls._handle_relationships(type_hint, name, purpose)
            fields[key] = (type_hint, _construct_field_info(elem, purpose, dto_field))

        return create_model(  # type:ignore[no-any-return,call-overload]
            name,
            __base__=cls,
            __module__=getattr(model, "__module__", __name__),
            __validators__=validators,
            __cls_kwargs__={"model": model},
            **fields,
        )

    @classmethod
    def _handle_relationships(cls, type_hint: Any, name: str, purpose: Purpose) -> Any:
        args = get_args(type_hint)
        if not args and not issubclass(type_hint, DeclarativeBase):
            return type_hint

        any_decl = any(issubclass(a, DeclarativeBase) for a in args)
        if args and not any_decl:
            return type_hint

        if args:
            origin_type = get_origin(type_hint)
            if origin_type is None:  # pragma: no cover
                raise RuntimeError("Unexpected `None` origin type.")
            origin_type = Union if origin_type is UnionType else origin_type
            inner_types = tuple(cls._handle_relationships(a, name, purpose) for a in args)
            return origin_type[inner_types]  # pyright:ignore

        type_hint = cls._factory(
            f"{name}_{type_hint.__name__}", type_hint, purpose=purpose, exclude=set()
        )
        return type_hint


def _construct_field_info(
    elem: Column | RelationshipProperty, purpose: Purpose, dto_field: DTOField
) -> FieldInfo:
    if dto_field.pydantic_field is not None:
        return dto_field.pydantic_field

    default = getattr(elem, "default", None)
    nullable = getattr(elem, "nullable", False)
    if purpose is Purpose.READ:
        return FieldInfo(...)
    if default is None:
        return FieldInfo(default=None) if nullable else FieldInfo(...)
    if default.is_scalar:
        return FieldInfo(default=default.arg)
    if default.is_callable:
        return FieldInfo(default_factory=lambda: default.arg({}))
    raise ValueError("Unexpected default type")


def _get_dto_field(elem: Column | RelationshipProperty) -> DTOField:
    return elem.info.get("dto", DTOField())


def _should_exclude_field(
    purpose: Purpose, elem: Column | RelationshipProperty, exclude: set[str], dto_attrib: DTOField
) -> bool:
    if elem.key in exclude:
        return True
    if dto_attrib.mark is Mark.PRIVATE:
        return True
    if purpose is Purpose.WRITE and dto_attrib.mark is Mark.READ_ONLY:
        return True
    return False


def _inspect_model(
    model: type[DeclarativeBase],
) -> tuple[ReadOnlyColumnCollection[str, Column], ReadOnlyProperties[RelationshipProperty]]:
    mapper = cast("Mapper", inspect(model))
    columns = mapper.columns
    relationships = mapper.relationships
    return columns, relationships


def _get_localns(model: type[DeclarativeBase]) -> dict[str, Any]:
    model_module = getmodule(model)
    return vars(model_module) if model_module is not None else {}



from __future__ import annotations

import sqlalchemy as sa

from quart_sqlalchemy import SQLAlchemy


def test_bind_key_default(db: SQLAlchemy) -> None:
    class User(db.Model):
        id = sa.Column(sa.Integer, primary_key=True)

    assert User.metadata is db.metadata


def test_metadata_per_bind(db: SQLAlchemy) -> None:
    class User(db.Model):
        __bind_key__ = "other"
        id = sa.Column(sa.Integer, primary_key=True)

    assert User.metadata is db.metadatas["other"]


def test_multiple_binds_same_table_name(db: SQLAlchemy) -> None:
    class UserA(db.Model):
        __tablename__ = "user"
        id = sa.Column(sa.Integer, primary_key=True)

    class UserB(db.Model):
        __bind_key__ = "other"
        __tablename__ = "user"
        id = sa.Column(sa.Integer, primary_key=True)

    assert UserA.metadata is db.metadata
    assert UserB.metadata is db.metadatas["other"]
    assert UserA.__table__.metadata is not UserB.__table__.metadata


def test_inherit_parent(db: SQLAlchemy) -> None:
    class User(db.Model):
        __bind_key__ = "auth"
        id = sa.Column(sa.Integer, primary_key=True)
        type = sa.Column(sa.String)
        __mapper_args__ = {"polymorphic_on": type, "polymorphic_identity": "user"}

    class Admin(User):
        id = sa.Column(sa.Integer, sa.ForeignKey(User.id), primary_key=True)
        __mapper_args__ = {"polymorphic_identity": "admin"}

    assert "admin" in db.metadatas["auth"].tables
    # inherits metadata, doesn't set it directly
    assert "metadata" not in Admin.__dict__


def test_inherit_abstract_parent(db: SQLAlchemy) -> None:
    class AbstractUser(db.Model):
        __abstract__ = True
        __bind_key__ = "auth"

    class User(AbstractUser):
        id = sa.Column(sa.Integer, primary_key=True)

    assert "user" in db.metadatas["auth"].tables
    assert "metadata" not in User.__dict__


def test_explicit_metadata(db: SQLAlchemy) -> None:
    other_metadata = sa.MetaData()

    class User(db.Model):
        __bind_key__ = "other"
        metadata = other_metadata
        id = sa.Column(sa.Integer, primary_key=True)

    assert User.__table__.metadata is other_metadata
    assert "other" not in db.metadatas


def test_explicit_table(db: SQLAlchemy) -> None:
    user_table = db.Table(
        "user",
        sa.Column("id", sa.Integer, primary_key=True),
        bind_key="auth",
    )

    class User(db.Model):
        __bind_key__ = "other"
        __table__ = user_table

    assert User.__table__.metadata is db.metadatas["auth"]
    assert "other" not in db.metadatas



# SPDX-FileCopyrightText: 2015 Eric Larson
#
# SPDX-License-Identifier: Apache-2.0

"""CacheControl import Interface.

Make it easy to import from cachecontrol without long namespaces.
"""
__author__ = "Eric Larson"
__email__ = "eric@ionrock.org"
__version__ = "0.12.10"

from .wrapper import CacheControl
from .adapter import CacheControlAdapter
from .controller import CacheController

import logging
logging.getLogger(__name__).addHandler(logging.NullHandler())



from datetime import datetime
from typing import Optional

from sqlalchemy import ForeignKey
from sqlalchemy import func
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import declared_attr
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import relationship


class HasId:
    """Mixin that adds an ``id`` column to a model.

    The column is an integer primary key that is auto-incremented.
    """

    id: Mapped[int] = mapped_column(primary_key=True)


class HasTimestamps:
    """Mixin that adds ``created_at`` and ``updated_at`` columns to a model.

    The columns are ``DateTime`` columns that default to the current time.
    """

    created_at: Mapped[datetime] = mapped_column(default=func.now(), nullable=False)
    updated_at: Mapped[datetime] = mapped_column(
        default=func.now(), nullable=False, onupdate=func.now()
    )


class SoftDelete:
    """Mixin that adds a ``deleted_at`` column to a model.

    The column is a ``DateTime`` column that defaults to ``NULL``.
    """

    deleted_at: Mapped[Optional[datetime]] = mapped_column(default=func.now())

    @property
    def is_active(self):
        return self.deleted_at is None


class HasVersion:
    """Mixin that adds a ``version`` column to a model.

    The column is an integer column that defaults to ``0``.
    """

    version: Mapped[int] = mapped_column(default=0)

    @declared_attr.directive
    def __mapper_args__(cls) -> Mapping[str, Any]:
        return dict(version_id_col=cls.version)


class HasVerified:
    """Mixin that adds a ``verified_at`` column to a model.

    The column is an datetime column that defaults to None.
    """

    verified_at: Mapped[Optional[datetime]] = mapped_column(default=None)

    @property
    def is_verified(self):
        return self.verified_at is not None

    def verify_model(self):
        self.verified_at = datetime.now()


class Consumable:
    """Mixin that adds a ``consumed_at`` column to a model.

    The column is an datetime column that defaults to None.
    """

    consumed_at: Mapped[Optional[datetime]] = mapped_column(default=None)

    @property
    def is_consumed(self):
        return self.consumed_at is not None

    def consume_model(self):
        self.consumed_at = datetime.now()


class Revokable:
    """Mixin that adds a ``revoked_at`` column to a model.

    The column is an datetime column that defaults to None.
    """

    revoked_at: Mapped[Optional[datetime]] = mapped_column(default=None)

    @property
    def is_revoked(self):
        return self.revoked_at is not None

    def revoke_model(self):
        self.revoked_at = datetime.now()


class HasExpiration:
    """Mixin that adds a ``expires_at`` column to a model.

    The column is an datetime column.
    """

    expires_at: Mapped[datetime] = mapped_column()

    @property
    def is_expired(self):
        return datetime.now() >= self.expires_at



from abc import ABC
from contextvars import ContextVar
from typing import Any
from typing import AsyncIterator
from typing import Mapping
from typing import Optional
from typing import Union

from aiopg.sa import SAConnection
from aiopg.sa.result import ResultProxy
from aiopg.sa.result import RowProxy
from aiopg.sa.transaction import Transaction as SATransaction
from sql_repo.repo.repk_old.base import AsyncBaseRepo
from sql_repo.repo.repk_old.base import AsyncQueryExecutor
from sql_repo.repo.repk_old.base import GenericIdModel
from sql_repo.repo.repk_old.impl import SqlAlchemyQuery


class AiopgRepository(AsyncBaseRepo[GenericIdModel], ABC):
    """
    Execute sql-queries, convert sql-row-dicts to/from pydantic models
    """

    def __init__(
        self, connection_or_context_var: Union[SAConnection, ContextVar[SAConnection]]
    ) -> None:
        self.connection_or_context_var = connection_or_context_var

    @property
    def _connection(self) -> SAConnection:
        if isinstance(self.connection_or_context_var, SAConnection):
            return self.connection_or_context_var
        else:
            return self.connection_or_context_var.get()

    @property
    def query_executor(self) -> AsyncQueryExecutor:
        return AiopgQueryExecutor(self._connection)


class AiopgQueryExecutor(AsyncQueryExecutor):
    def __init__(self, connection: SAConnection) -> None:
        self._connection = connection

    async def fetch_one(self, query: SqlAlchemyQuery, **sa_params: Any) -> Optional[Mapping]:
        rows: ResultProxy = await self._connection.execute(query, **sa_params)
        row: RowProxy = await rows.first()
        return row

    async def fetch_all(self, query: SqlAlchemyQuery, **sa_params: Any) -> AsyncIterator[Mapping]:
        return await self._connection.execute(query, **sa_params)

    async def fetch_val(self, query: SqlAlchemyQuery, **sa_params: Any) -> Any:
        return await self._connection.scalar(query, **sa_params)

    async def insert(self, query: SqlAlchemyQuery, **sa_params: Any) -> Mapping:
        rows = await self._connection.execute(query, **sa_params)
        row = await rows.first()
        return row

    async def insert_many(self, query: SqlAlchemyQuery, **sa_params: Any) -> AsyncIterator[Mapping]:
        return await self._connection.execute(query, **sa_params)

    async def update(self, query: SqlAlchemyQuery, **sa_params: Any) -> None:
        return await self._connection.execute(query, **sa_params)

    async def delete(self, query: SqlAlchemyQuery, **sa_params: Any) -> None:
        await self._connection.execute(query, **sa_params)

    def execute_in_transaction(self) -> SATransaction:
        return self._connection.begin()



from __future__ import annotations

import os
import typing as t
from weakref import WeakKeyDictionary

import sqlalchemy
import sqlalchemy.orm
from quart import abort
from quart import current_app
from quart import has_app_context
from quart import Quart
from sqlalchemy import engine_from_config
from sqlalchemy import event
from sqlalchemy import MetaData
from sqlalchemy import Table
from sqlalchemy.engine import Engine
from sqlalchemy.engine import make_url
from sqlalchemy.engine import URL
from sqlalchemy.exc import MultipleResultsFound
from sqlalchemy.exc import NoResultFound
from sqlalchemy.exc import UnboundExecutionError
from sqlalchemy.orm import declarative_base
from sqlalchemy.orm import DeclarativeMeta
from sqlalchemy.orm import dynamic_loader
from sqlalchemy.orm import relationship
from sqlalchemy.orm import RelationshipProperty
from sqlalchemy.orm import scoped_session
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import QueuePool
from sqlalchemy.pool import StaticPool
from sqlalchemy.sql import Select

from .model import _QueryProperty
from .model import DefaultMeta
from .model import Model
from .pagination import Pagination
from .pagination import SelectPagination
from .query import Query
from .session import _app_ctx_id
from .session import Session


class SQLAlchemy:
    """Integrates SQLAlchemy with Quart. This handles setting up one or more engines,
    associating tables and models with specific engines, and cleaning up connections and
    sessions after each request.

    Only the engine configuration is specific to each application, other things like
    the model, table, metadata, and session are shared for all applications using that
    extension instance. Call :meth:`init_app` to configure the extension on an
    application.

    After creating the extension, create model classes by subclassing :attr:`Model`, and
    table classes with :attr:`Table`. These can be accessed before :meth:`init_app` is
    called, making it possible to define the models separately from the application.

    Accessing :attr:`session` and :attr:`engine` requires an active Quart application
    context. This includes methods like :meth:`create_all` which use the engine.

    This class also provides access to names in SQLAlchemy's ``sqlalchemy`` and
    ``sqlalchemy.orm`` modules. For example, you can use ``db.Column`` and
    ``db.relationship`` instead of importing ``sqlalchemy.Column`` and
    ``sqlalchemy.orm.relationship``. This can be convenient when defining models.

    :param app: Call :meth:`init_app` on this Quart application now.
    :param metadata: Use this as the default :class:`sqlalchemy.schema.MetaData`. Useful
        for setting a naming convention.
    :param session_options: Arguments used by :attr:`session` to create each session
        instance. A ``scopefunc`` key will be passed to the scoped session, not the
        session instance. See :class:`sqlalchemy.orm.sessionmaker` for a list of
        arguments.
    :param query_class: Use this as the default query class for models and dynamic
        relationships. The query interface is considered legacy in SQLAlchemy.
    :param model_class: Use this as the model base class when creating the declarative
        model class :attr:`Model`. Can also be a fully created declarative model class
        for further customization.
    :param engine_options: Default arguments used when creating every engine. These are
        lower precedence than application config. See :func:`sqlalchemy.create_engine`
        for a list of arguments.
    :param add_models_to_shell: Add the ``db`` instance and all model classes to
        ``quart shell``.

    .. versionchanged:: 3.0
        An active Quart application context is always required to access ``session`` and
        ``engine``.

    .. versionchanged:: 3.0
        Separate ``metadata`` are used for each bind key.

    .. versionchanged:: 3.0
        The ``engine_options`` parameter is applied as defaults before per-engine
        configuration.

    .. versionchanged:: 3.0
        The session class can be customized in ``session_options``.

    .. versionchanged:: 3.0
        Added the ``add_models_to_shell`` parameter.

    .. versionchanged:: 3.0
        Engines are created when calling ``init_app`` rather than the first time they
        are accessed.

    .. versionchanged:: 3.0
        All parameters except ``app`` are keyword-only.

    .. versionchanged:: 3.0
        The extension instance is stored directly as ``app.extensions["sqlalchemy"]``.

    .. versionchanged:: 3.0
        Setup methods are renamed with a leading underscore. They are considered
        internal interfaces which may change at any time.

    .. versionchanged:: 3.0
        Removed the ``use_native_unicode`` parameter and config.

    .. versionchanged:: 2.4
        Added the ``engine_options`` parameter.

    .. versionchanged:: 2.1
        Added the ``metadata``, ``query_class``, and ``model_class`` parameters.

    .. versionchanged:: 2.1
        Use the same query class across ``session``, ``Model.query`` and
        ``Query``.

    .. versionchanged:: 0.16
        ``scopefunc`` is accepted in ``session_options``.

    .. versionchanged:: 0.10
        Added the ``session_options`` parameter.
    """

    def __init__(
        self,
        app: Quart | None = None,
        *,
        metadata: MetaData | None = None,
        session_options: dict[str, t.Any] | None = None,
        query_class: t.Type[Query] = Query,
        model_class: t.Type[Model] | DeclarativeMeta = Model,
        engine_options: dict[str, t.Any] | None = None,
        add_models_to_shell: bool = True,
    ):
        if session_options is None:
            session_options = {}

        self.Query = query_class
        """The default query class used by ``Model.query`` and ``lazy="dynamic"``
        relationships.

        .. warning::
            The query interface is considered legacy in SQLAlchemy.

        Customize this by passing the ``query_class`` parameter to the extension.
        """

        self.session = self._make_scoped_session(session_options)
        """A :class:`sqlalchemy.orm.scoping.scoped_session` that creates instances of
        :class:`.Session` scoped to the current Quart application context. The session
        will be removed, returning the engine connection to the pool, when the
        application context exits.

        Customize this by passing ``session_options`` to the extension.

        This requires that a Quart application context is active.

        .. versionchanged:: 3.0
            The session is scoped to the current app context.
        """

        self.metadatas: dict[str | None, MetaData] = {}
        """Map of bind keys to :class:`sqlalchemy.schema.MetaData` instances. The
        ``None`` key refers to the default metadata, and is available as
        :attr:`metadata`.

        Customize the default metadata by passing the ``metadata`` parameter to the
        extension. This can be used to set a naming convention. When metadata for
        another bind key is created, it copies the default's naming convention.

        .. versionadded:: 3.0
        """

        if metadata is not None:
            metadata.info["bind_key"] = None
            self.metadatas[None] = metadata

        self.Table = self._make_table_class()
        """A :class:`sqlalchemy.schema.Table` class that chooses a metadata
        automatically.

        Unlike the base ``Table``, the ``metadata`` argument is not required. If it is
        not given, it is selected based on the ``bind_key`` argument.

        :param bind_key: Used to select a different metadata.
        :param args: Arguments passed to the base class. These are typically the table's
            name, columns, and constraints.
        :param kwargs: Arguments passed to the base class.

        .. versionchanged:: 3.0
            This is a subclass of SQLAlchemy's ``Table`` rather than a function.
        """

        self.Model = self._make_declarative_base(model_class)
        """A SQLAlchemy declarative model class. Subclass this to define database
        models.

        If a model does not set ``__tablename__``, it will be generated by converting
        the class name from ``CamelCase`` to ``snake_case``. It will not be generated
        if the model looks like it uses single-table inheritance.

        If a model or parent class sets ``__bind_key__``, it will use that metadata and
        database engine. Otherwise, it will use the default :attr:`metadata` and
        :attr:`engine`. This is ignored if the model sets ``metadata`` or ``__table__``.

        Customize this by subclassing :class:`.Model` and passing the ``model_class``
        parameter to the extension. A fully created declarative model class can be
        passed as well, to use a custom metaclass.
        """

        if engine_options is None:
            engine_options = {}

        self._engine_options = engine_options
        self._app_engines: WeakKeyDictionary[Quart, dict[str | None, Engine]]
        self._app_engines = WeakKeyDictionary()
        self._add_models_to_shell = add_models_to_shell

        if app is not None:
            self.init_app(app)

    def __repr__(self) -> str:
        if not has_app_context():
            return f"<{type(self).__name__}>"

        message = f"{type(self).__name__} {self.engine.url}"

        if len(self.engines) > 1:
            message = f"{message} +{len(self.engines) - 1}"

        return f"<{message}>"

    def init_app(self, app: Quart) -> None:
        """Initialize a Quart application for use with this extension instance. This
        must be called before accessing the database engine or session with the app.

        This sets default configuration values, then configures the extension on the
        application and creates the engines for each bind key. Therefore, this must be
        called after the application has been configured. Changes to application config
        after this call will not be reflected.

        The following keys from ``app.config`` are used:

        - :data:`.SQLALCHEMY_DATABASE_URI`
        - :data:`.SQLALCHEMY_ENGINE_OPTIONS`
        - :data:`.SQLALCHEMY_ECHO`
        - :data:`.SQLALCHEMY_BINDS`
        - :data:`.SQLALCHEMY_RECORD_QUERIES`
        - :data:`.SQLALCHEMY_TRACK_MODIFICATIONS`

        :param app: The Quart application to initialize.
        """
        app.extensions["sqlalchemy"] = self
        app.teardown_appcontext(self._teardown_session)

        if self._add_models_to_shell:
            from .cli import add_models_to_shell

            app.shell_context_processor(add_models_to_shell)

        basic_uri: str | URL | None = app.config.setdefault("SQLALCHEMY_DATABASE_URI", None)
        basic_engine_options = self._engine_options.copy()
        basic_engine_options.update(app.config.setdefault("SQLALCHEMY_ENGINE_OPTIONS", {}))
        echo: bool = app.config.setdefault("SQLALCHEMY_ECHO", False)
        config_binds: dict[str | None, str | URL | dict[str, t.Any]] = app.config.setdefault(
            "SQLALCHEMY_BINDS", {}
        )
        engine_options: dict[str | None, dict[str, t.Any]] = {}

        # Build the engine config for each bind key.
        for key, value in config_binds.items():
            engine_options[key] = self._engine_options.copy()

            if isinstance(value, (str, URL)):
                engine_options[key]["url"] = value
            else:
                engine_options[key].update(value)

        # Build the engine config for the default bind key.
        if basic_uri is not None:
            basic_engine_options["url"] = basic_uri

        if "url" in basic_engine_options:
            engine_options.setdefault(None, {}).update(basic_engine_options)

        if not engine_options:
            raise RuntimeError(
                "Either 'SQLALCHEMY_DATABASE_URI' or 'SQLALCHEMY_BINDS' must be set."
            )

        engines = self._app_engines.setdefault(app, {})

        # Dispose existing engines in case init_app is called again.
        if engines:
            for engine in engines.values():
                engine.dispose()

            engines.clear()

        # Create the metadata and engine for each bind key.
        for key, options in engine_options.items():
            self._make_metadata(key)
            options.setdefault("echo", echo)
            options.setdefault("echo_pool", echo)
            self._apply_driver_defaults(options, app)
            engines[key] = self._make_engine(key, options, app)

        if app.config.setdefault("SQLALCHEMY_RECORD_QUERIES", False):
            from . import record_queries

            for engine in engines.values():
                record_queries._listen(engine)

        if app.config.setdefault("SQLALCHEMY_TRACK_MODIFICATIONS", False):
            from . import track_modifications

            track_modifications._listen(self.session)

    def _make_scoped_session(self, options: dict[str, t.Any]) -> scoped_session:
        """Create a :class:`sqlalchemy.orm.scoping.scoped_session` around the factory
        from :meth:`_make_session_factory`. The result is available as :attr:`session`.

        The scope function can be customized using the ``scopefunc`` key in the
        ``session_options`` parameter to the extension. By default it uses the current
        thread or greenlet id.

        This method is used for internal setup. Its signature may change at any time.

        :meta private:

        :param options: The ``session_options`` parameter from ``__init__``. Keyword
            arguments passed to the session factory. A ``scopefunc`` key is popped.

        .. versionchanged:: 3.0
            The session is scoped to the current app context.

        .. versionchanged:: 3.0
            Renamed from ``create_scoped_session``, this method is internal.
        """
        scope = options.pop("scopefunc", _app_ctx_id)
        factory = self._make_session_factory(options)
        return scoped_session(factory, scope)

    def _make_session_factory(
        self, options: dict[str, t.Any]
    ) -> sessionmaker[Session]:  # type: ignore[type-var]
        """Create the SQLAlchemy :class:`sqlalchemy.orm.sessionmaker` used by
        :meth:`_make_scoped_session`.

        To customize, pass the ``session_options`` parameter to :class:`SQLAlchemy`. To
        customize the session class, subclass :class:`.Session` and pass it as the
        ``class_`` key.

        This method is used for internal setup. Its signature may change at any time.

        :meta private:

        :param options: The ``session_options`` parameter from ``__init__``. Keyword
            arguments passed to the session factory.

        .. versionchanged:: 3.0
            The session class can be customized.

        .. versionchanged:: 3.0
            Renamed from ``create_session``, this method is internal.
        """
        options.setdefault("class_", Session)
        options.setdefault("query_cls", self.Query)
        return sessionmaker(db=self, **options)

    def _teardown_session(self, exc: BaseException | None) -> None:
        """Remove the current session at the end of the request.

        :meta private:

        .. versionadded:: 3.0
        """
        self.session.remove()

    def _make_metadata(self, bind_key: str | None) -> MetaData:
        """Get or create a :class:`sqlalchemy.schema.MetaData` for the given bind key.

        This method is used for internal setup. Its signature may change at any time.

        :meta private:

        :param bind_key: The name of the metadata being created.

        .. versionadded:: 3.0
        """
        if bind_key in self.metadatas:
            return self.metadatas[bind_key]

        if bind_key is not None:
            # Copy the naming convention from the default metadata.
            naming_convention = self._make_metadata(None).naming_convention
        else:
            naming_convention = None

        # Set the bind key in info to be used by session.get_bind.
        metadata = MetaData(naming_convention=naming_convention, info={"bind_key": bind_key})
        self.metadatas[bind_key] = metadata
        return metadata

    def _make_table_class(self) -> t.Type[Table]:
        """Create a SQLAlchemy :class:`sqlalchemy.schema.Table` class that chooses a
        metadata automatically based on the ``bind_key``. The result is available as
        :attr:`Table`.

        This method is used for internal setup. Its signature may change at any time.

        :meta private:

        .. versionadded:: 3.0
        """

        class Table(sqlalchemy.Table):
            def __new__(cls, *args: t.Any, bind_key: str | None = None, **kwargs: t.Any) -> Table:
                # If a metadata arg is passed, go directly to the base Table. Also do
                # this for no args so the correct error is shown.
                if not args or (len(args) >= 2 and isinstance(args[1], MetaData)):
                    return super().__new__(cls, *args, **kwargs)

                metadata = self._make_metadata(bind_key)
                return super().__new__(cls, args[0], metadata, *args[1:], **kwargs)

        return Table

    def _make_declarative_base(self, model: t.Type[Model] | DeclarativeMeta) -> t.Type[t.Any]:
        """Create a SQLAlchemy declarative model class. The result is available as
        :attr:`Model`.

        To customize, subclass :class:`.Model` and pass it as ``model_class`` to
        :class:`SQLAlchemy`. To customize at the metaclass level, pass an already
        created declarative model class as ``model_class``.

        This method is used for internal setup. Its signature may change at any time.

        :meta private:

        :param model: A model base class, or an already created declarative model class.

        .. versionchanged:: 3.0
            Renamed with a leading underscore, this method is internal.

        .. versionchanged:: 2.3
            ``model`` can be an already created declarative model class.
        """
        if not isinstance(model, DeclarativeMeta):
            metadata = self._make_metadata(None)
            model = declarative_base(
                metadata=metadata, cls=model, name="Model", metaclass=DefaultMeta
            )

        if None not in self.metadatas:
            # Use the model's metadata as the default metadata.
            model.metadata.info["bind_key"] = None  # type: ignore[union-attr]
            self.metadatas[None] = model.metadata  # type: ignore[union-attr]
        else:
            # Use the passed in default metadata as the model's metadata.
            model.metadata = self.metadatas[None]  # type: ignore[union-attr]

        model.query_class = self.Query
        model.query = _QueryProperty()
        model.__fsa__ = self
        return model

    def _apply_driver_defaults(self, options: dict[str, t.Any], app: Quart) -> None:
        """Apply driver-specific configuration to an engine.

        SQLite in-memory databases use ``StaticPool`` and disable ``check_same_thread``.
        File paths are relative to the app's :attr:`~quart.Quart.instance_path`,
        which is created if it doesn't exist.

        MySQL sets ``charset="utf8mb4"``, and ``pool_timeout`` defaults to 2 hours.

        This method is used for internal setup. Its signature may change at any time.

        :meta private:

        :param options: Arguments passed to the engine.
        :param app: The application that the engine configuration belongs to.

        .. versionchanged:: 3.0
            SQLite paths are relative to ``app.instance_path``. It does not use
            ``NullPool`` if ``pool_size`` is 0. Driver-level URIs are supported.

        .. versionchanged:: 3.0
            MySQL sets ``charset="utf8mb4". It does not set ``pool_size`` to 10. It
            does not set ``pool_recycle`` if not using a queue pool.

        .. versionchanged:: 3.0
            Renamed from ``apply_driver_hacks``, this method is internal. It does not
            return anything.

        .. versionchanged:: 2.5
            Returns ``(sa_url, options)``.
        """
        url = make_url(options["url"])

        if url.drivername in {"sqlite", "sqlite+pysqlite"}:
            if url.database is None or url.database in {"", ":memory:"}:
                options["poolclass"] = StaticPool

                if "connect_args" not in options:
                    options["connect_args"] = {}

                options["connect_args"]["check_same_thread"] = False
            else:
                # the url might look like sqlite:///file:path?uri=true
                is_uri = url.query.get("uri", False)

                if is_uri:
                    db_str = url.database[5:]
                else:
                    db_str = url.database

                if not os.path.isabs(db_str):
                    os.makedirs(app.instance_path, exist_ok=True)
                    db_str = os.path.join(app.instance_path, db_str)

                    if is_uri:
                        db_str = f"file:{db_str}"

                    options["url"] = url.set(database=db_str)
        elif url.drivername.startswith("mysql"):
            # set queue defaults only when using queue pool
            if "pool_class" not in options or options["pool_class"] is QueuePool:
                options.setdefault("pool_recycle", 7200)

            if "charset" not in url.query:
                options["url"] = url.update_query_dict({"charset": "utf8mb4"})

    def _make_engine(self, bind_key: str | None, options: dict[str, t.Any], app: Quart) -> Engine:
        """Create the :class:`sqlalchemy.engine.Engine` for the given bind key and app.

        To customize, use :data:`.SQLALCHEMY_ENGINE_OPTIONS` or
        :data:`.SQLALCHEMY_BINDS` config. Pass ``engine_options`` to :class:`SQLAlchemy`
        to set defaults for all engines.

        This method is used for internal setup. Its signature may change at any time.

        :meta private:

        :param bind_key: The name of the engine being created.
        :param options: Arguments passed to the engine.
        :param app: The application that the engine configuration belongs to.

        .. versionchanged:: 3.0
            Renamed from ``create_engine``, this method is internal.
        """
        return engine_from_config(options, prefix="")

    @property
    def metadata(self) -> MetaData:
        """The default metadata used by :attr:`Model` and :attr:`Table` if no bind key
        is set.
        """
        return self.metadatas[None]

    @property
    def engines(self) -> t.Mapping[str | None, Engine]:
        """Map of bind keys to :class:`sqlalchemy.engine.Engine` instances for current
        application. The ``None`` key refers to the default engine, and is available as
        :attr:`engine`.

        To customize, set the :data:`.SQLALCHEMY_BINDS` config, and set defaults by
        passing the ``engine_options`` parameter to the extension.

        This requires that a Quart application context is active.

        .. versionadded:: 3.0
        """
        app = current_app._get_current_object()
        return self._app_engines[app]

    # def get_app(self):
    #     if hasattr(self, 'app'):
    #         return self.app
    #     return current_app._get_current_object()

    @property
    def engine(self) -> Engine:
        """The default :class:`~sqlalchemy.engine.Engine` for the current application,
        used by :attr:`session` if the :attr:`Model` or :attr:`Table` being queried does
        not set a bind key.

        To customize, set the :data:`.SQLALCHEMY_ENGINE_OPTIONS` config, and set
        defaults by passing the ``engine_options`` parameter to the extension.

        This requires that a Quart application context is active.
        """
        return self.engines[None]

    def get_or_404(
        self, entity: t.Type[t.Any], ident: t.Any, *, description: str | None = None
    ) -> t.Any:
        """Like :meth:`session.get() <sqlalchemy.orm.Session.get>` but aborts with a
        ``404 Not Found`` error instead of returning ``None``.

        :param entity: The model class to query.
        :param ident: The primary key to query.
        :param description: A custom message to show on the error page.

        .. versionadded:: 3.0
        """
        value = self.session.get(entity, ident)

        if value is None:
            abort(404, description=description)

        return value

    def first_or_404(self, statement: Select, *, description: str | None = None) -> t.Any:
        """Like :meth:`Result.scalar() <sqlalchemy.engine.Result.scalar>`, but aborts
        with a ``404 Not Found`` error instead of returning ``None``.

        :param statement: The ``select`` statement to execute.
        :param description: A custom message to show on the error page.

        .. versionadded:: 3.0
        """
        value = self.session.execute(statement).scalar()

        if value is None:
            abort(404, description=description)

        return value

    def one_or_404(self, statement: Select, *, description: str | None = None) -> t.Any:
        """Like :meth:`Result.scalar_one() <sqlalchemy.engine.Result.scalar_one>`,
        but aborts with a ``404 Not Found`` error instead of raising ``NoResultFound``
        or ``MultipleResultsFound``.

        :param statement: The ``select`` statement to execute.
        :param description: A custom message to show on the error page.

        .. versionadded:: 3.0
        """
        try:
            return self.session.execute(statement).scalar_one()
        except (NoResultFound, MultipleResultsFound):
            abort(404, description=description)

    def paginate(
        self,
        select: Select,
        *,
        page: int | None = None,
        per_page: int | None = None,
        max_per_page: int | None = None,
        error_out: bool = True,
        count: bool = True,
    ) -> Pagination:
        """Apply an offset and limit to a select statment based on the current page and
        number of items per page, returning a :class:`.Pagination` object.

        The statement should select a model class, like ``select(User)``. This applies
        ``unique()`` and ``scalars()`` modifiers to the result, so compound selects will
        not return the expected results.

        :param select: The ``select`` statement to paginate.
        :param page: The current page, used to calculate the offset. Defaults to the
            ``page`` query arg during a request, or 1 otherwise.
        :param per_page: The maximum number of items on a page, used to calculate the
            offset and limit. Defaults to the ``per_page`` query arg during a request,
            or 20 otherwise.
        :param max_per_page: The maximum allowed value for ``per_page``, to limit a
            user-provided value. Use ``None`` for no limit. Defaults to 100.
        :param error_out: Abort with a ``404 Not Found`` error if no items are returned
            and ``page`` is not 1, or if ``page`` or ``per_page`` is less than 1, or if
            either are not ints.
        :param count: Calculate the total number of values by issuing an extra count
            query. For very complex queries this may be inaccurate or slow, so it can be
            disabled and set manually if necessary.

        .. versionchanged:: 3.0
            The ``count`` query is more efficient.

        .. versionadded:: 3.0
        """
        return SelectPagination(
            select=select,
            session=self.session(),
            page=page,
            per_page=per_page,
            max_per_page=max_per_page,
            error_out=error_out,
            count=count,
        )

    def _call_for_binds(self, bind_key: str | None | list[str | None], op_name: str) -> None:
        """Call a method on each metadata.

        :meta private:

        :param bind_key: A bind key or list of keys. Defaults to all binds.
        :param op_name: The name of the method to call.

        .. versionchanged:: 3.0
            Renamed from ``_execute_for_all_tables``.
        """
        if bind_key == "__all__":
            keys: list[str | None] = list(self.metadatas)
        elif bind_key is None or isinstance(bind_key, str):
            keys = [bind_key]
        else:
            keys = bind_key

        for key in keys:
            try:
                engine = self.engines[key]
            except KeyError:
                message = f"Bind key '{key}' is not in 'SQLALCHEMY_BINDS' config."

                if key is None:
                    message = f"'SQLALCHEMY_DATABASE_URI' config is not set. {message}"

                raise UnboundExecutionError(message) from None

            metadata = self.metadatas[key]
            getattr(metadata, op_name)(bind=engine)

    def create_all(self, bind_key: str | None | list[str | None] = "__all__") -> None:
        """Create tables that do not exist in the database by calling
        ``metadata.create_all()`` for all or some bind keys. This does not
        update existing tables, use a migration library for that.

        This requires that a Quart application context is active.

        :param bind_key: A bind key or list of keys to create the tables for. Defaults
            to all binds.

        .. versionchanged:: 3.0
            Renamed the ``bind`` parameter to ``bind_key``. Removed the ``app``
            parameter.

        .. versionchanged:: 0.12
            Added the ``bind`` and ``app`` parameters.
        """
        self._call_for_binds(bind_key, "create_all")

    def drop_all(self, bind_key: str | None | list[str | None] = "__all__") -> None:
        """Drop tables by calling ``metadata.drop_all()`` for all or some bind keys.

        This requires that a Quart application context is active.

        :param bind_key: A bind key or list of keys to drop the tables from. Defaults to
            all binds.

        .. versionchanged:: 3.0
            Renamed the ``bind`` parameter to ``bind_key``. Removed the ``app``
            parameter.

        .. versionchanged:: 0.12
            Added the ``bind`` and ``app`` parameters.
        """
        self._call_for_binds(bind_key, "drop_all")

    def reflect(self, bind_key: str | None | list[str | None] = "__all__") -> None:
        """Load table definitions from the database by calling ``metadata.reflect()``
        for all or some bind keys.

        This requires that a Quart application context is active.

        :param bind_key: A bind key or list of keys to reflect the tables from. Defaults
            to all binds.

        .. versionchanged:: 3.0
            Renamed the ``bind`` parameter to ``bind_key``. Removed the ``app``
            parameter.

        .. versionchanged:: 0.12
            Added the ``bind`` and ``app`` parameters.
        """
        self._call_for_binds(bind_key, "reflect")

    def _set_rel_query(self, kwargs: dict[str, t.Any]) -> None:
        """Apply the extension's :attr:`Query` class as the default for relationships
        and backrefs.

        :meta private:
        """
        kwargs.setdefault("query_class", self.Query)

        if "backref" in kwargs:
            backref = kwargs["backref"]

            if isinstance(backref, str):
                backref = (backref, {})

            backref[1].setdefault("query_class", self.Query)

    def relationship(self, *args: t.Any, **kwargs: t.Any) -> RelationshipProperty[t.Any]:
        """A :func:`sqlalchemy.orm.relationship` that applies this extension's
        :attr:`Query` class for dynamic relationships and backrefs.

        .. versionchanged:: 3.0
            The :attr:`Query` class is set on ``backref``.
        """
        self._set_rel_query(kwargs)
        return relationship(*args, **kwargs)

    def dynamic_loader(self, argument: t.Any, **kwargs: t.Any) -> RelationshipProperty[t.Any]:
        """A :func:`sqlalchemy.orm.dynamic_loader` that applies this extension's
        :attr:`Query` class for relationships and backrefs.

        .. versionchanged:: 3.0
            The :attr:`Query` class is set on ``backref``.
        """
        self._set_rel_query(kwargs)
        return dynamic_loader(argument, **kwargs)

    def _relation(self, *args: t.Any, **kwargs: t.Any) -> RelationshipProperty[t.Any]:
        """A :func:`sqlalchemy.orm.relationship` that applies this extension's
        :attr:`Query` class for dynamic relationships and backrefs.

        SQLAlchemy 2.0 removes this name, use ``relationship`` instead.

        :meta private:

        .. versionchanged:: 3.0
            The :attr:`Query` class is set on ``backref``.
        """
        self._set_rel_query(kwargs)
        return relation(*args, **kwargs)

    def __getattr__(self, name: str) -> t.Any:
        if name == "relation":
            return self._relation

        if name == "event":
            return event

        if name.startswith("_"):
            raise AttributeError(name)

        for mod in (sqlalchemy, sqlalchemy.orm):
            if hasattr(mod, name):
                return getattr(mod, name)

        raise AttributeError(name)



from __future__ import annotations

import pytest
import sqlalchemy as sa
import sqlalchemy.orm
from quart import Quart

from quart_sqlalchemy import SQLAlchemy
from quart_sqlalchemy.model import DefaultMeta
from quart_sqlalchemy.model import Model


def test_default_metadata(db: SQLAlchemy) -> None:
    assert db.metadata is db.metadatas[None]
    assert db.metadata.info["bind_key"] is None
    assert db.Model.metadata is db.metadata


def test_custom_metadata() -> None:
    metadata = sa.MetaData()
    db = SQLAlchemy(metadata=metadata)
    assert db.metadata is metadata
    assert db.metadata.info["bind_key"] is None
    assert db.Model.metadata is db.metadata


def test_metadata_from_custom_model() -> None:
    base = sa.orm.declarative_base(cls=Model, metaclass=DefaultMeta)
    metadata = base.metadata  # type: ignore[attr-defined]
    db = SQLAlchemy(model_class=base)
    assert db.Model.metadata is metadata
    assert db.Model.metadata is db.metadata


def test_custom_metadata_overrides_custom_model() -> None:
    base = sa.orm.declarative_base(cls=Model, metaclass=DefaultMeta)
    metadata = sa.MetaData()
    db = SQLAlchemy(model_class=base, metadata=metadata)
    assert db.Model.metadata is metadata
    assert db.Model.metadata is db.metadata


def test_metadata_per_bind(app: Quart) -> None:
    app.config["SQLALCHEMY_BINDS"] = {"a": "sqlite://"}
    db = SQLAlchemy(app)
    assert db.metadatas["a"] is not db.metadata
    assert db.metadatas["a"].info["bind_key"] == "a"


def test_copy_naming_convention(app: Quart) -> None:
    app.config["SQLALCHEMY_BINDS"] = {"a": "sqlite://"}
    db = SQLAlchemy(
        app, metadata=sa.MetaData(naming_convention={"pk": "spk_%(table_name)s"})
    )
    assert db.metadata.naming_convention["pk"] == "spk_%(table_name)s"
    assert db.metadatas["a"].naming_convention == db.metadata.naming_convention


async def test_create_drop_all(app: Quart) -> None:
    async with app.app_context():
        app.config["SQLALCHEMY_BINDS"] = {"a": "sqlite://"}
        db = SQLAlchemy(app)

        class User(db.Model):
            id = sa.Column(sa.Integer, primary_key=True)

        class Post(db.Model):
            __bind_key__ = "a"
            id = sa.Column(sa.Integer, primary_key=True)

        with pytest.raises(sa.exc.OperationalError):
            db.session.execute(sa.select(User)).scalars()

        with pytest.raises(sa.exc.OperationalError):
            db.session.execute(sa.select(Post)).scalars()

        db.create_all()
        db.session.execute(sa.select(User)).scalars()
        db.session.execute(sa.select(Post)).scalars()
        db.drop_all()

        with pytest.raises(sa.exc.OperationalError):
            db.session.execute(sa.select(User)).scalars()

        with pytest.raises(sa.exc.OperationalError):
            db.session.execute(sa.select(Post)).scalars()


@pytest.mark.parametrize("bind_key", ["a", ["a"]])
async def test_create_key_spec(app: Quart, bind_key: str | list[str | None]) -> None:
    async with app.app_context():
        app.config["SQLALCHEMY_BINDS"] = {"a": "sqlite://"}
        db = SQLAlchemy(app)

        class User(db.Model):
            id = sa.Column(sa.Integer, primary_key=True)

        class Post(db.Model):
            __bind_key__ = "a"
            id = sa.Column(sa.Integer, primary_key=True)

        db.create_all(bind_key=bind_key)
        db.session.execute(sa.select(Post)).scalars()

        with pytest.raises(sa.exc.OperationalError):
            db.session.execute(sa.select(User)).scalars()



async def test_reflect(app: Quart) -> None:
    async with app.app_context():
        app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite:///user.db"
        app.config["SQLALCHEMY_BINDS"] = {"post": "sqlite:///post.db"}
        db = SQLAlchemy(app)
        db.Table("user", sa.Column("id", sa.Integer, primary_key=True))
        db.Table("post", sa.Column("id", sa.Integer, primary_key=True), bind_key="post")
        db.create_all()

        db = SQLAlchemy(app)
        assert not db.metadata.tables
        db.reflect()
        assert "user" in db.metadata.tables
        assert "post" in db.metadatas["post"].tables



from quart import Blueprint
from quart import flash
from quart import g
from quart import redirect
from quart import render_template
from quart import request
from quart import url_for
from werkzeug.exceptions import abort

from quartr import db
from quartr.auth.views import login_required
from quartr.blog.models import Post

bp = Blueprint("blog", __name__)


@bp.route("/")
def index():
    """Show all the posts, most recent first."""
    select = db.select(Post).order_by(Post.created.desc())
    posts = db.session.execute(select).scalars()
    return render_template("blog/index.html", posts=posts)


def get_post(id, check_author=True):
    """Get a post and its author by id.

    Checks that the id exists and optionally that the current user is
    the author.

    :param id: id of post to get
    :param check_author: require the current user to be the author
    :return: the post with author information
    :raise 404: if a post with the given id doesn't exist
    :raise 403: if the current user isn't the author
    """
    post = db.get_or_404(Post, id, description=f"Post id {id} doesn't exist.")

    if check_author and post.author != g.user:
        abort(403)

    return post


@bp.route("/create", methods=("GET", "POST"))
@login_required
def create():
    """Create a new post for the current user."""
    if request.method == "POST":
        title = request.form["title"]
        body = request.form["body"]
        error = None

        if not title:
            error = "Title is required."

        if error is not None:
            flash(error)
        else:
            db.session.add(Post(title=title, body=body, author=g.user))
            db.session.commit()
            return redirect(url_for("blog.index"))

    return render_template("blog/create.html")


@bp.route("/<int:id>/update", methods=("GET", "POST"))
@login_required
def update(id):
    """Update a post if the current user is the author."""
    post = get_post(id)

    if request.method == "POST":
        title = request.form["title"]
        body = request.form["body"]
        error = None

        if not title:
            error = "Title is required."

        if error is not None:
            flash(error)
        else:
            post.title = title
            post.body = body
            db.session.commit()
            return redirect(url_for("blog.index"))

    return render_template("blog/update.html", post=post)


@bp.route("/<int:id>/delete", methods=("POST",))
@login_required
def delete(id):
    """Delete a post.

    Ensures that the post exists and that the logged in user is the
    author of the post.
    """
    post = get_post(id)
    db.session.delete(post)
    db.session.commit()
    return redirect(url_for("blog.index"))



from __future__ import annotations

import typing as t

import sqlalchemy as sa
import sqlalchemy.event
import sqlalchemy.orm
from quart import current_app
from quart import has_app_context
from quart.signals import Namespace  # type: ignore[attr-defined]

if t.TYPE_CHECKING:
    from .session import Session

_signals = Namespace()

models_committed = _signals.signal("models-committed")
"""This Blinker signal is sent after the session is committed if there were changed
models in the session.

The sender is the application that emitted the changes. The receiver is passed the
``changes`` argument with a list of tuples in the form ``(instance, operation)``.
The operations are ``"insert"``, ``"update"``, and ``"delete"``.
"""

before_models_committed = _signals.signal("before-models-committed")
"""This signal works exactly like :data:`models_committed` but is emitted before the
commit takes place.
"""


def _listen(session: sa.orm.scoped_session) -> None:
    sa.event.listen(session, "before_flush", _record_ops, named=True)
    sa.event.listen(session, "before_commit", _record_ops, named=True)
    sa.event.listen(session, "before_commit", _before_commit)
    sa.event.listen(session, "after_commit", _after_commit)
    sa.event.listen(session, "after_rollback", _after_rollback)


def _record_ops(session: Session, **kwargs: t.Any) -> None:
    if not has_app_context():
        return

    if not current_app.config["SQLALCHEMY_TRACK_MODIFICATIONS"]:
        return

    for targets, operation in (
        (session.new, "insert"),
        (session.dirty, "update"),
        (session.deleted, "delete"),
    ):
        for target in targets:
            state = sa.inspect(target)
            key = state.identity_key if state.has_identity else id(target)
            session._model_changes[key] = (target, operation)


def _before_commit(session: Session) -> None:
    if not has_app_context():
        return

    app = current_app._get_current_object()  # type: ignore[attr-defined]

    if not app.config["SQLALCHEMY_TRACK_MODIFICATIONS"]:
        return

    if session._model_changes:
        changes = list(session._model_changes.values())
        before_models_committed.send(app, changes=changes)


def _after_commit(session: Session) -> None:
    if not has_app_context():
        return

    app = current_app._get_current_object()  # type: ignore[attr-defined]

    if not app.config["SQLALCHEMY_TRACK_MODIFICATIONS"]:
        return

    if session._model_changes:
        changes = list(session._model_changes.values())
        models_committed.send(app, changes=changes)
        session._model_changes.clear()


def _after_rollback(session: Session) -> None:
    session._model_changes.clear()



from typing import Any

import boto3
from mypy_boto3_kms.client import KMSClient
from pydantic import BaseModel as PydanticBaseModel
from pydantic import ConstrainedStr
from pydantic import SecretStr

from two.crypto_util import decrypt
from two.crypto_util import encrypt
from two.crypto_util import get_kms_client


key_id = "91d9a9c8-0651-4ded-8c93-e5ed9e06d8f1"
kms = boto3.client("kms")


class BaseModel(PydanticBaseModel):
    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        validate_all = True
        orm_mode = True
        use_enum_values = True

    @classmethod
    def _get_value(cls, v: Any, *args: Any, **kwargs: Any) -> Any:
        """Hack to enable smarter serialization behavior.

        This wrapper will check for a `__serialize__` instance method defined on the custom field
        class and use that for serialization if it exists.  Then it will check the type against
        the json_encoders defined in the model config and use the first one that matches.  Failing
        that, it will fallback to default behvior.

        By doing this before rather than after, we can even ensure correct serialization
        of custom field types that inherit from python scalars such as str, int, etc.

        This work is inspired by: https://github.com/pydantic/pydantic/issues/951
        """
        if hasattr(v, "__serialize__"):
            return v.__serialize__()
        for type_, converter in cls.__config__.json_encoders.items():
            if isinstance(v, type_):
                return converter(v)

        return super()._get_value(v, *args, **kwargs)


class EncryptedSecret(ConstrainedStr):
    @classmethod
    def validate(cls, value, config):
        key_id, kms = cls._get_kms_args(config)
        decrypted = decrypt(key_id, value, kms)
        return ConstrainedStr(decrypted)

    @classmethod
    def _get_kms_args(cls, config):
        key_id = getattr(config, "kms_key_id", None)
        kms = getattr(config, "kms_client", None)
        if key_id is None or kms is None:
            raise ValueError(
                f"kms_key_id and kms_client must be set in Config to use {cls.__name__}"
            )
        if callable(kms):
            kms = kms()
        return key_id, kms

    def encrypt(self, key_id: str, kms: KMSClient):
        value = encrypt(key_id, self, kms)
        return EncryptedSecret(value)

    def __repr__(self):
        return f"{type(self).__name__}({super().__repr__()})"

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(  # pragma: no cover
            examples=["jNkGJXh4LTcIfkWfUpatlUpEWXfJvU71"],
        )


# class Example(BaseModel):
#     class Config:
#         kms_key_id = "some_key"
#         kms_client = get_kms_client()

#     secret: EncryptedSecret = EncryptedSecret(
#         "AQICAHi4b5SNvuUHTVjg/flY91hbzyWOMWLx7tvWqwyLKT1V2gF738bj8MEr0Oea7WYMBELnAAAAZDBiBgkqhkiG9w0BBwagVTBTAgEAME4GCSqGSIb3DQEHATAeBglghkgBZQMEAS4wEQQMtmdJmtU8GSChFKmFAgEQgCFSDhugVHHiywNaFUfhb7odnsdMjxKH2BwO5EwDnDCJG/Y="
#     )



"""Application ORM configuration."""
from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Any
from typing import Callable
from typing import Iterable
from typing import Optional
from typing import TypeVar
from uuid import UUID
from uuid import uuid4

from pydantic.fields import FieldInfo
from sqlalchemy import MetaData
from sqlalchemy.dialects import postgresql as pg
from sqlalchemy.event import listens_for
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import declared_attr
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import registry
from sqlalchemy.orm import Session
from sqlalchemy_utils import JSONType
from sqlalchemy_utils import UUIDType


# from starlite_saqlalchemy import dto, settings

BaseT = TypeVar("BaseT", bound="Base")

# DTO_KEY = "dto"
"""The key we use to reference `dto.DTOField` in the SQLAlchemy info dict."""

convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}
"""
Templates for automated constraint name generation.
"""


# class Mark(str, Enum):
#     """For marking column definitions on the domain models.

#     Example:
#     ```python
#     class Model(Base):
#         ...
#         updated_at: Mapped[datetime] = mapped_column(info={"dto": Mark.READ_ONLY})
#     ```
#     """

#     READ_ONLY = "read-only"
#     """To mark a field that can be read, but not updated by clients."""
#     PRIVATE = "private"
#     """To mark a field that can neither be read or updated by clients."""


@listens_for(Session, "before_flush")
def touch_updated_timestamp(session: Session, *_: Any) -> None:
    """Set timestamp on update.

    Called from SQLAlchemy's
    [`before_flush`][sqlalchemy.orm.SessionEvents.before_flush] event to bump the `updated`
    timestamp on modified instances.

    Args:
        session: The sync [`Session`][sqlalchemy.orm.Session] instance that underlies the async
            session.
    """
    for instance in session.dirty:
        instance.updated = datetime.now()


# @dataclass
# class DTOField:
#     """For configuring DTO behavior on SQLAlchemy model fields."""

#     mark: Mark | None = None
#     """Mark the field as read-only, or private."""
#     pydantic_type: Any | None = None
#     """Override the field type on the pydantic model for this attribute."""
#     pydantic_field: FieldInfo | None = None
#     """If provided, used for the pydantic model for this attribute."""
#     validators: Iterable[Callable[[Any], Any]] | None = None
#     """Single argument callables that are defined on the DTO as validators for the field."""


class Base(DeclarativeBase):
    """Base for all SQLAlchemy declarative models."""

    registry = registry(
        metadata=MetaData(naming_convention=convention),
        type_annotation_map={UUID: UUIDType, dict: JSONType},
    )

    id: Mapped[int] = mapped_column(
        default=uuid4, primary_key=True, info={"dto": DTOField(mark=Mark.READ_ONLY)}
    )
    """Primary key column."""
    created: Mapped[datetime] = mapped_column(
        default=datetime.now, info={"dto": DTOField(mark=Mark.READ_ONLY)}
    )
    """Date/time of instance creation."""
    updated: Mapped[datetime] = mapped_column(
        default=datetime.now, info={"dto": DTOField(mark=Mark.READ_ONLY)}
    )
    """Date/time of instance update."""

    # noinspection PyMethodParameters
    @declared_attr.directive
    def __tablename__(cls) -> str:  # pylint: disable=no-self-argument
        """Infer table name from class name."""
        return cls.__name__.lower()



import pytest
from quart import g
from quart import session

from quartr import db
from quartr.auth.models import User


def test_register(client, app):
    # test that viewing the page renders without template errors
    assert client.get("/auth/register").status_code == 200

    # test that successful registration redirects to the login page
    response = client.post("/auth/register", data={"username": "a", "password": "a"})
    assert response.headers["Location"] == "/auth/login"

    # test that the user was inserted into the database
    with app.app_context():
        select = db.select(User).filter_by(username="a")
        user = db.session.execute(select).scalar()
        assert user is not None


def test_user_password(app):
    user = User(username="a", password="a")
    assert user.password_hash != "a"
    assert user.check_password("a")


@pytest.mark.parametrize(
    ("username", "password", "message"),
    (
        ("", "", "Username is required."),
        ("a", "", "Password is required."),
        ("test", "test", "already registered"),
    ),
)
def test_register_validate_input(client, username, password, message):
    response = client.post(
        "/auth/register", data={"username": username, "password": password}
    )
    assert message in response.text


def test_login(client, auth):
    # test that viewing the page renders without template errors
    assert client.get("/auth/login").status_code == 200

    # test that successful login redirects to the index page
    response = auth.login()
    assert response.headers["Location"] == "/"

    # login request set the user_id in the session
    # check that the user is loaded from the session
    with client:
        client.get("/")
        assert session["user_id"] == 1
        assert g.user.username == "test"


@pytest.mark.parametrize(
    ("username", "password", "message"),
    (("a", "test", "Incorrect username."), ("test", "a", "Incorrect password.")),
)
def test_login_validate_input(auth, username, password, message):
    response = auth.login(username, password)
    assert message in response.text


def test_logout(client, auth):
    auth.login()

    with client:
        auth.logout()
        assert "user_id" not in session



# from __future__ import annotations

# from dataclasses import dataclass
from dataclasses import field
from typing import List
from typing import Optional

import anyio
from pydantic.dataclasses import dataclass
from quart import Quart
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import select
from sqlalchemy import String
from sqlalchemy import Table
from sqlalchemy.orm import registry
from sqlalchemy.orm import relationship
from sqlalchemy.orm import selectinload

from quart_sqlalchemy import SQLAlchemy


registry = registry()

app = Quart(__name__)
app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite://"
db = SQLAlchemy(app, metadata=registry.metadata)


@dataclass
class User:
    id: Optional[int] = field(init=False, default=None)
    name: Optional[str] = None
    fullname: Optional[str] = None
    nickname: Optional[str] = None
    addresses: List["Address"] = field(default_factory=list)


@dataclass
class Address:
    id: Optional[int] = field(init=False, default=None)
    user_id: Optional[int] = field(init=False, default=None)
    email_address: Optional[str] = None


User.__pydantic_model__.update_forward_refs()

user = Table(
    "user",
    registry.metadata,
    Column("id", Integer, primary_key=True),
    Column("name", String(50)),
    Column("fullname", String(50)),
    Column("nickname", String(12)),
)

address = Table(
    "address",
    registry.metadata,
    Column("id", Integer, primary_key=True),
    Column("user_id", Integer, ForeignKey("user.id")),
    Column("email_address", String(50)),
)

registry.map_imperatively(
    User,
    user,
    properties={
        "addresses": relationship(Address, backref="user", order_by=address.c.id),
    },
)

registry.map_imperatively(Address, address)


async def main():
    async with app.app_context():
        db.create_all()

        user = User(
            "joe",
            "Joe Black",
            "Joe",
            [Address(email_address="joe@joe.com")],
        )
        with db.session() as session:
            session.add(user)
            session.commit()
            session.refresh(user)
            session.expunge(user)

        # print(user)
        import pdb

        pdb.set_trace()

        with db.session() as session:
            statement = select(User).where(User.name == "joe").options(selectinload(User.addresses))
            model = session.execute(statement).scalars().one()
            assert model.id == user.id


anyio.run(main)



from abc import ABC
from contextlib import asynccontextmanager
from typing import Any
from typing import cast
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

from sql_repo.repo.repk_old.base import AsyncBaseRepo
from sql_repo.repo.repk_old.base import Columns
from sql_repo.repo.repk_old.base import Created
from sql_repo.repo.repk_old.base import GenericIdModel
from sql_repo.repo.repk_old.impl import Filters
from sqlalchemy.sql.elements import BinaryExpression


class FakeRepo(AsyncBaseRepo[GenericIdModel], ABC):
    def __init__(self) -> None:
        self.entities: List[GenericIdModel] = []
        self.id_counter = 1

    async def first(
        self, *filters: BinaryExpression, orders: Optional[Columns] = None
    ) -> Optional[GenericIdModel]:
        return next(iter(self.entities), None)

    async def get_by_ids(self, entity_ids: Sequence[int]) -> List[GenericIdModel]:
        return [entity for entity in self.entities if entity.id in entity_ids]

    async def get_by_id(self, entity_id: int) -> Optional[GenericIdModel]:
        return next((entity for entity in self.entities if entity.id == entity_id), None)

    async def get_or_create(
        self, filters: Filters = None, defaults: Dict = None
    ) -> Tuple[GenericIdModel, Created]:
        raise NotImplementedError()

    async def get_all(
        self, filters: Filters = None, orders: Columns = None
    ) -> List[GenericIdModel]:
        return self.entities

    async def get_all_ids(
        self, filters: Sequence[BinaryExpression] = None, orders: Columns = None
    ) -> Sequence[int]:
        return [cast(int, entity.id) for entity in self.entities]

    async def exists(self, *filters: BinaryExpression) -> bool:
        raise NotImplementedError()

    async def insert(self, entity: GenericIdModel) -> GenericIdModel:
        entity.id = self.id_counter
        self.id_counter += 1
        self.entities.append(entity)
        return entity

    async def insert_many(self, entities: List[GenericIdModel]) -> List[GenericIdModel]:
        return [await self.insert(entity) for entity in entities]

    async def update(self, entity: GenericIdModel) -> GenericIdModel:
        return entity

    async def update_partial(self, entity: GenericIdModel, **updated_values: Any) -> GenericIdModel:
        for field, value in updated_values.items():
            setattr(entity, field, value)
        return entity

    async def update_many(self, entities: List[GenericIdModel]) -> List[GenericIdModel]:
        return entities

    async def delete(self, *filters: Optional[BinaryExpression]) -> None:
        raise NotImplementedError()

    async def delete_by_id(self, entity_id: int) -> None:
        self.entities = [entity for entity in self.entities if entity.id != entity_id]

    async def delete_by_ids(self, entity_ids: Sequence[int]) -> None:
        self.entities = [entity for entity in self.entities if entity.id not in entity_ids]

    @asynccontextmanager
    def execute_in_transaction(self) -> Any:
        yield None



from __future__ import annotations

import sqlalchemy as sa
from quart import Quart

from quart_sqlalchemy import SQLAlchemy
from quart_sqlalchemy.record_queries import get_recorded_queries


async def test_query_info(app: Quart) -> None:
    async with app.app_context():
        app.config["SQLALCHEMY_RECORD_QUERIES"] = True
        db = SQLAlchemy(app)

        class Example(db.Model):
            id = sa.Column(sa.Integer, primary_key=True)

        db.create_all()
        db.session.execute(sa.select(Example).filter(Example.id < 5)).scalars()
        info = get_recorded_queries()[-1]
        assert info.statement is not None
        assert "SELECT" in info.statement
        assert "FROM example" in info.statement
        assert info.parameters[0][0] == 5
        assert info.duration == info.end_time - info.start_time
        assert "tests/test_record_queries.py:" in info.location
        assert "(test_query_info)" in info.location



from __future__ import annotations

import typing as t

import pytest
import sqlalchemy as sa
import sqlalchemy.orm
from quart import Quart

from quart_sqlalchemy import SQLAlchemy
from quart_sqlalchemy.model import DefaultMeta
from quart_sqlalchemy.model import Model


def test_default_model_class(db: SQLAlchemy) -> None:
    assert db.Model.query_class is db.Query
    assert db.Model.metadata is db.metadata
    assert issubclass(db.Model, Model)
    assert isinstance(db.Model, DefaultMeta)


def test_custom_model_class(app: Quart) -> None:
    class CustomModel(Model):
        pass

    db = SQLAlchemy(app, model_class=CustomModel)
    assert issubclass(db.Model, CustomModel)
    assert isinstance(db.Model, DefaultMeta)


@pytest.mark.parametrize("base", [Model, object])
async def test_custom_declarative_class(app: Quart, base: t.Any) -> None:
    class CustomMeta(DefaultMeta):
        pass

    async with app.app_context():
        CustomModel = sa.orm.declarative_base(cls=base, name="Model", metaclass=CustomMeta)
        db = SQLAlchemy(app, model_class=CustomModel)
        assert db.Model is CustomModel
        assert db.Model.query_class is db.Query
        assert "query" in db.Model.__dict__


async def test_model_repr(app: Quart, db: SQLAlchemy) -> None:
    class User(db.Model):
        id = sa.Column(sa.Integer, primary_key=True)

    async with app.app_context():
        db.create_all()
        user = User()
        assert repr(user) == f"<User (transient {id(user)})>"
        db.session.add(user)
        assert repr(user) == f"<User (pending {id(user)})>"
        db.session.flush()
        assert repr(user) == f"<User {user.id}>"



from __future__ import annotations

from typing import Generator, Any
import warnings

import pytest
import sqlalchemy as sa
from quart import Quart
from werkzeug.exceptions import NotFound

from quart_sqlalchemy import SQLAlchemy
from quart_sqlalchemy.query import Query


@pytest.fixture(autouse=True)
def ignore_query_warning() -> Generator[None, None, None]:
    if hasattr(sa.exc, "LegacyAPIWarning"):
        with warnings.catch_warnings():
            exc = sa.exc.LegacyAPIWarning  # type: ignore[attr-defined]
            warnings.simplefilter("ignore", exc)
            yield
    else:
        yield


async def test_get_or_404(app: Quart, db: SQLAlchemy, Todo: Any) -> None:
    async with app.app_context():
        item = Todo()
        db.session.add(item)
        db.session.commit()
        assert Todo.query.get_or_404(1) is item

        with pytest.raises(NotFound):
            Todo.query.get_or_404(2)


async def test_first_or_404(app: Quart, db: SQLAlchemy, Todo: Any) -> None:
    async with app.app_context():
        db.session.add(Todo(title="a"))
        db.session.commit()
        assert Todo.query.filter_by(title="a").first_or_404().title == "a"

        with pytest.raises(NotFound):
            Todo.query.filter_by(title="b").first_or_404()


async def test_one_or_404(app: Quart, db: SQLAlchemy, Todo: Any) -> None:
    async with app.app_context():
        db.session.add(Todo(title="a"))
        db.session.add(Todo(title="b"))
        db.session.add(Todo(title="b"))
        db.session.commit()
        assert Todo.query.filter_by(title="a").one_or_404().title == "a"

        with pytest.raises(NotFound):
            # MultipleResultsFound
            Todo.query.filter_by(title="b").one_or_404()

        with pytest.raises(NotFound):
            # NoResultFound
            Todo.query.filter_by(title="c").one_or_404()


async def test_paginate(app: Quart, db: SQLAlchemy, Todo: Any) -> None:
    async with app.app_context():
        db.session.add_all(Todo() for _ in range(150))
        db.session.commit()
        p = Todo.query.paginate()
        assert p.total == 150
        assert len(p.items) == 20
        p2 = p.next()
        assert p2.page == 2
        assert p2.total == 150


async def test_default_query_class(app: Quart, db: SQLAlchemy) -> None:
    class Parent(db.Model):
        id = sa.Column(sa.Integer, primary_key=True)
        children1 = db.relationship("Child", backref="parent1", lazy="dynamic")

    class Child(db.Model):
        id = sa.Column(sa.Integer, primary_key=True)
        parent_id = sa.Column(sa.ForeignKey(Parent.id))
        parent2 = db.relationship(
            Parent,
            backref=db.backref("children2", lazy="dynamic", viewonly=True),
            viewonly=True,
        )

    async with app.app_context():
        p = Parent()
        assert type(Parent.query) is Query
        assert isinstance(p.children1, Query)
        assert isinstance(p.children2, Query)
        assert isinstance(db.session.query(Child), Query)


async def test_custom_query_class(app: Quart) -> None:
    class CustomQuery(Query):
        pass

    async with app.app_context():
        db = SQLAlchemy(app, query_class=CustomQuery)

        class Parent(db.Model):
            id = sa.Column(sa.Integer, primary_key=True)
            children1 = db.relationship("Child", backref="parent1", lazy="dynamic")

        class Child(db.Model):
            id = sa.Column(sa.Integer, primary_key=True)
            parent_id = sa.Column(sa.ForeignKey(Parent.id))
            parent2 = db.relationship(
                Parent,
                backref=db.backref("children2", lazy="dynamic", viewonly=True),
                viewonly=True,
            )

        p = Parent()
        assert type(Parent.query) is CustomQuery
        assert isinstance(p.children1, CustomQuery)
        assert isinstance(p.children2, CustomQuery)
        assert isinstance(db.session.query(Child), CustomQuery)



from datetime import datetime

import pytest
from werkzeug.security import generate_password_hash

from quart import create_app
from quart import db
from quart import init_db
from quart.auth.models import User
from quart.blog.models import Post

_user1_pass = generate_password_hash("test")
_user2_pass = generate_password_hash("other")


@pytest.fixture
def app():
    """Create and configure a new app instance for each test."""
    # create the app with common test config
    app = create_app({"TESTING": True, "SQLALCHEMY_DATABASE_URI": "sqlite:///:memory:"})

    # create the database and load test data
    # set _password to pre-generated hashes, since hashing for each test is slow
    with app.app_context():
        init_db()
        user = User(username="test", password_hash=_user1_pass)
        db.session.add_all(
            (
                user,
                User(username="other", password_hash=_user2_pass),
                Post(
                    title="test title",
                    body="test\nbody",
                    author=user,
                    created=datetime(2018, 1, 1),
                ),
            )
        )
        db.session.commit()

    yield app


@pytest.fixture
def client(app):
    """A test client for the app."""
    return app.test_client()


@pytest.fixture
def runner(app):
    """A test runner for the app's Click commands."""
    return app.test_cli_runner()


class AuthActions:
    def __init__(self, client):
        self._client = client

    def login(self, username="test", password="test"):
        return self._client.post(
            "/auth/login", data={"username": username, "password": password}
        )

    def logout(self):
        return self._client.get("/auth/logout")


@pytest.fixture
def auth(client):
    return AuthActions(client)



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from typing import List
from typing import Optional

from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import registry
from sqlalchemy.orm import relationship


mapper_registry = registry()


@mapper_registry.mapped
@dataclass
class User:
    __tablename__ = "user"
    __sa_dataclass_metadata_key__ = "sa"

    id: int = field(init=False, metadata={"sa": Column(Integer, primary_key=True)})
    name: str = field(default=None, metadata={"sa": Column(String(50))})
    fullname: str = field(default=None, metadata={"sa": Column(String(50))})
    nickname: str = field(default=None, metadata={"sa": Column(String(12))})

    addresses: List[Address] = field(default_factory=list, metadata={"sa": relationship("Address")})


@mapper_registry.mapped
@dataclass
class Address:
    __tablename__ = "address"
    __sa_dataclass_metadata_key__ = "sa"

    id: int = field(init=False, metadata={"sa": Column(Integer, primary_key=True)})
    user_id: int = field(init=False, metadata={"sa": Column(ForeignKey("user.id"))})
    email_address: Optional[str] = field(default=None, metadata={"sa": Column(String(50))})


engine = create_engine("sqlite:///:memory:")
mapper_registry.metadata.create_all(bind=engine)

user = User(name="Michael")
user.addresses = [
    Address(email_address="foo@bar.com"),
    Address(email_address="bar@foo.com"),
]

with Session(engine, expire_on_commit=False) as session, session.begin():
    session.add(user)

print(user)
print(astuple(user))
print(json.dumps(asdict(user), indent=2))



from __future__ import annotations

import re
import typing as t

import sqlalchemy as sa
import sqlalchemy.orm

from .query import Query

if t.TYPE_CHECKING:
    from .extension import SQLAlchemy


class _QueryProperty:
    """A class property that creates a query object for a model.

    :meta private:
    """

    def __get__(self, obj: Model | None, cls: t.Type[Model]) -> Query:
        return cls.query_class(
            cls, session=cls.__fsa__.session()  # type: ignore[arg-type]
        )


class Model:
    """The base class of the :attr:`.SQLAlchemy.Model` declarative model class.

    To define models, subclass :attr:`db.Model <.SQLAlchemy.Model>`, not this. To
    customize ``db.Model``, subclass this and pass it as ``model_class`` to
    :class:`.SQLAlchemy`. To customize ``db.Model`` at the metaclass level, pass an
    already created declarative model class as ``model_class``.
    """

    __fsa__: t.ClassVar[SQLAlchemy]
    """Internal reference to the extension object.

    :meta private:
    """

    query_class: t.ClassVar[t.Type[Query]] = Query
    """Query class used by :attr:`query`. Defaults to :attr:`.SQLAlchemy.Query`, which
    defaults to :class:`.Query`.
    """

    query: t.ClassVar[Query] = _QueryProperty()  # type: ignore[assignment]
    """A SQLAlchemy query for a model. Equivalent to ``db.session.query(Model)``. Can be
    customized per-model by overriding :attr:`query_class`.

    .. warning::
        The query interface is considered legacy in SQLAlchemy. Prefer using
        ``session.execute(select())`` instead.
    """

    def __repr__(self) -> str:
        state = sa.inspect(self)

        if state.transient:
            pk = f"(transient {id(self)})"
        elif state.pending:
            pk = f"(pending {id(self)})"
        else:
            pk = ", ".join(map(str, state.identity))

        return f"<{type(self).__name__} {pk}>"


class BindMetaMixin(type):
    """Metaclass mixin that sets a model's ``metadata`` based on its ``__bind_key__``.

    If the model sets ``metadata`` or ``__table__`` directly, ``__bind_key__`` is
    ignored. If the ``metadata`` is the same as the parent model, it will not be set
    directly on the child model.
    """

    __fsa__: SQLAlchemy
    metadata: sa.MetaData

    def __init__(
        cls, name: str, bases: tuple[type, ...], d: dict[str, t.Any], **kwargs: t.Any
    ) -> None:
        if not ("metadata" in cls.__dict__ or "__table__" in cls.__dict__):
            bind_key = getattr(cls, "__bind_key__", None)
            parent_metadata = getattr(cls, "metadata", None)
            metadata = cls.__fsa__._make_metadata(bind_key)

            if metadata is not parent_metadata:
                cls.metadata = metadata

        super().__init__(name, bases, d, **kwargs)


class NameMetaMixin(type):
    """Metaclass mixin that sets a model's ``__tablename__`` by converting the
    ``CamelCase`` class name to ``snake_case``. A name is set for non-abstract models
    that do not otherwise define ``__tablename__``. If a model does not define a primary
    key, it will not generate a name or ``__table__``, for single-table inheritance.
    """

    metadata: sa.MetaData
    __tablename__: str
    __table__: sa.Table

    def __init__(
        cls, name: str, bases: tuple[type, ...], d: dict[str, t.Any], **kwargs: t.Any
    ) -> None:
        if should_set_tablename(cls):
            cls.__tablename__ = camel_to_snake_case(cls.__name__)

        super().__init__(name, bases, d, **kwargs)

        # __table_cls__ has run. If no table was created, use the parent table.
        if (
            "__tablename__" not in cls.__dict__
            and "__table__" in cls.__dict__
            and cls.__dict__["__table__"] is None
        ):
            del cls.__table__

    def __table_cls__(cls, *args: t.Any, **kwargs: t.Any) -> sa.Table | None:
        """This is called by SQLAlchemy during mapper setup. It determines the final
        table object that the model will use.

        If no primary key is found, that indicates single-table inheritance, so no table
        will be created and ``__tablename__`` will be unset.
        """
        schema = kwargs.get("schema")

        if schema is None:
            key = args[0]
        else:
            key = f"{schema}.{args[0]}"

        # Check if a table with this name already exists. Allows reflected tables to be
        # applied to models by name.
        if key in cls.metadata.tables:
            return sa.Table(*args, **kwargs)

        # If a primary key is found, create a table for joined-table inheritance.
        for arg in args:
            if (isinstance(arg, sa.Column) and arg.primary_key) or isinstance(
                arg, sa.PrimaryKeyConstraint
            ):
                return sa.Table(*args, **kwargs)

        # If no base classes define a table, return one that's missing a primary key
        # so SQLAlchemy shows the correct error.
        for base in cls.__mro__[1:-1]:
            if "__table__" in base.__dict__:
                break
        else:
            return sa.Table(*args, **kwargs)

        # Single-table inheritance, use the parent table name. __init__ will unset
        # __table__ based on this.
        if "__tablename__" in cls.__dict__:
            del cls.__tablename__

        return None


def should_set_tablename(cls: type) -> bool:
    """Determine whether ``__tablename__`` should be generated for a model.

    -   If no class in the MRO sets a name, one should be generated.
    -   If a declared attr is found, it should be used instead.
    -   If a name is found, it should be used if the class is a mixin, otherwise one
        should be generated.
    -   Abstract models should not have one generated.

    Later, ``__table_cls__`` will determine if the model looks like single or
    joined-table inheritance. If no primary key is found, the name will be unset.
    """
    if cls.__dict__.get("__abstract__", False) or not any(
        isinstance(b, sa.orm.DeclarativeMeta) for b in cls.__mro__[1:]
    ):
        return False

    for base in cls.__mro__:
        if "__tablename__" not in base.__dict__:
            continue

        if isinstance(base.__dict__["__tablename__"], sa.orm.declared_attr):
            return False

        return not (
            base is cls
            or base.__dict__.get("__abstract__", False)
            or not isinstance(base, sa.orm.DeclarativeMeta)
        )

    return True


def camel_to_snake_case(name: str) -> str:
    """Convert a ``CamelCase`` name to ``snake_case``."""
    name = re.sub(r"((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))", r"_\1", name)
    return name.lower().lstrip("_")


class DefaultMeta(BindMetaMixin, NameMetaMixin, sa.orm.DeclarativeMeta):
    """SQLAlchemy declarative metaclass that provides ``__bind_key__`` and
    ``__tablename__`` support.
    """



from datetime import datetime
from datetime import timezone

from quart import url_for

from quartr import db
from quartr.auth.models import User


def now_utc():
    return datetime.now(timezone.utc)


class Post(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    author_id = db.Column(db.ForeignKey(User.id), nullable=False)
    created = db.Column(db.DateTime, nullable=False, default=now_utc)
    title = db.Column(db.String, nullable=False)
    body = db.Column(db.String, nullable=False)

    # User object backed by author_id
    # lazy="joined" means the user is returned with the post in one query
    author = db.relationship(User, lazy="joined", back_populates="posts")

    @property
    def update_url(self):
        return url_for("blog.update", id=self.id)

    @property
    def delete_url(self):
        return url_for("blog.delete", id=self.id)



from quartr import create_app


def test_config():
    """Test create_app without passing test config."""
    assert not create_app().testing
    assert create_app({"TESTING": True}).testing


def test_db_url_environ(monkeypatch):
    """Test DATABASE_URL environment variable."""
    monkeypatch.setenv("DATABASE_URL", "sqlite:///environ")
    app = create_app()
    assert app.config["SQLALCHEMY_DATABASE_URI"] == "sqlite:///environ"


def test_init_db_command(runner, monkeypatch):
    called = False

    def fake_init_db():
        nonlocal called
        called = True

    monkeypatch.setattr("quartr.init_db", fake_init_db)
    result = runner.invoke(args=["init-db"])
    assert "Initialized" in result.output
    assert called



from __future__ import annotations

import inspect
import typing as t

import pytest
import sqlalchemy as sa
import sqlalchemy.orm

from quart_sqlalchemy import SQLAlchemy
from quart_sqlalchemy.model import camel_to_snake_case


@pytest.mark.parametrize(
    ("name", "expect"),
    [
        ("CamelCase", "camel_case"),
        ("Snake_case", "snake_case"),
        ("HTMLLayout", "html_layout"),
        ("LayoutHTML", "layout_html"),
        ("HTTP2Request", "http2_request"),
        ("ShoppingCartSession", "shopping_cart_session"),
        ("ABC", "abc"),
        ("PreABC", "pre_abc"),
        ("ABCPost", "abc_post"),
        ("PreABCPost", "pre_abc_post"),
        ("HTTP2RequestSession", "http2_request_session"),
        ("UserST4", "user_st4"),
        (
            "HTTP2ClientType3EncoderParametersSSE",
            "http2_client_type3_encoder_parameters_sse",
        ),
        (
            "LONGName4TestingCamelCase2snake_caseXYZ",
            "long_name4_testing_camel_case2snake_case_xyz",
        ),
        ("FooBarSSE2", "foo_bar_sse2"),
        ("AlarmMessageSS2SignalTransformer", "alarm_message_ss2_signal_transformer"),
        ("AstV2Node", "ast_v2_node"),
        ("HTTPResponseCodeXYZ", "http_response_code_xyz"),
        ("get2HTTPResponse123Code", "get2_http_response123_code"),
        # ("getHTTPresponseCode", "get_htt_presponse_code"),
        # ("__test__Method", "test___method"),
    ],
)
def test_camel_to_snake_case(name: str, expect: str) -> None:
    assert camel_to_snake_case(name) == expect


def test_name(db: SQLAlchemy) -> None:
    class FOOBar(db.Model):
        id = sa.Column(sa.Integer, primary_key=True)

    class BazBar(db.Model):
        id = sa.Column(sa.Integer, primary_key=True)

    class Ham(db.Model):
        __tablename__ = "spam"
        id = sa.Column(sa.Integer, primary_key=True)

    assert FOOBar.__tablename__ == "foo_bar"
    assert BazBar.__tablename__ == "baz_bar"
    assert Ham.__tablename__ == "spam"


def test_single_name(db: SQLAlchemy) -> None:
    """Single table inheritance should not set a new name."""

    class Duck(db.Model):
        id = sa.Column(sa.Integer, primary_key=True)

    class Mallard(Duck):
        pass

    assert "__tablename__" not in Mallard.__dict__
    assert Mallard.__tablename__ == "duck"


def test_joined_name(db: SQLAlchemy) -> None:
    """Model has a separate primary key; it should set a new name."""

    class Duck(db.Model):
        id = sa.Column(sa.Integer, primary_key=True)

    class Donald(Duck):
        id = sa.Column(sa.Integer, sa.ForeignKey(Duck.id), primary_key=True)

    assert Donald.__tablename__ == "donald"


def test_mixin_id(db: SQLAlchemy) -> None:
    """Primary key provided by mixin should still allow model to set
    tablename.
    """

    class Base:
        id = sa.Column(sa.Integer, primary_key=True)

    class Duck(Base, db.Model):
        pass

    assert not hasattr(Base, "__tablename__")
    assert Duck.__tablename__ == "duck"


def test_mixin_attr(db: SQLAlchemy) -> None:
    """A declared attr tablename will be used down multiple levels of
    inheritance.
    """

    class Mixin:
        @sa.orm.declared_attr
        def __tablename__(cls) -> str:  # noqa: B902
            return cls.__name__.upper()  # type: ignore[attr-defined,no-any-return]

    class Bird(Mixin, db.Model):
        id = sa.Column(sa.Integer, primary_key=True)

    class Duck(Bird):
        # object reference
        id = sa.Column(sa.Integer, sa.ForeignKey(Bird.id), primary_key=True)

    class Mallard(Duck):
        # string reference
        id = sa.Column(sa.Integer, sa.ForeignKey("DUCK.id"), primary_key=True)

    assert Bird.__tablename__ == "BIRD"
    assert Duck.__tablename__ == "DUCK"
    assert Mallard.__tablename__ == "MALLARD"


def test_abstract_name(db: SQLAlchemy) -> None:
    """Abstract model should not set a name. Subclass should set a name."""

    class Base(db.Model):
        __abstract__ = True
        id = sa.Column(sa.Integer, primary_key=True)

    class Duck(Base):
        pass

    assert "__tablename__" not in Base.__dict__
    assert Duck.__tablename__ == "duck"


def test_complex_inheritance(db: SQLAlchemy) -> None:
    """Joined table inheritance, but the new primary key is provided by a
    mixin, not directly on the class.
    """

    class Duck(db.Model):
        id = sa.Column(sa.Integer, primary_key=True)

    class IdMixin:
        @sa.orm.declared_attr
        def id(cls):  # type: ignore[no-untyped-def]  # noqa: B902
            return sa.Column(sa.Integer, sa.ForeignKey(Duck.id), primary_key=True)

    class RubberDuck(IdMixin, Duck):  # type: ignore[misc]
        pass

    assert RubberDuck.__tablename__ == "rubber_duck"


def test_manual_name(db: SQLAlchemy) -> None:
    """Setting a manual name prevents generation for the immediate model. A
    name is generated for joined but not single-table inheritance.
    """

    class Duck(db.Model):
        __tablename__ = "DUCK"
        id = sa.Column(sa.Integer, primary_key=True)
        type = sa.Column(sa.String)

        __mapper_args__ = {"polymorphic_on": type}

    class Daffy(Duck):
        id = sa.Column(sa.Integer, sa.ForeignKey(Duck.id), primary_key=True)

        __mapper_args__ = {"polymorphic_identity": "Tower"}  # type: ignore[dict-item]

    class Donald(Duck):
        __mapper_args__ = {"polymorphic_identity": "Mouse"}  # type: ignore[dict-item]

    assert Duck.__tablename__ == "DUCK"
    assert Daffy.__tablename__ == "daffy"
    assert "__tablename__" not in Donald.__dict__
    assert Donald.__tablename__ == "DUCK"


def test_primary_constraint(db: SQLAlchemy) -> None:
    """Primary key will be picked up from table args."""

    class Duck(db.Model):
        id = sa.Column(sa.Integer)

        __table_args__ = (sa.PrimaryKeyConstraint(id),)

    assert Duck.__table__ is not None
    assert Duck.__tablename__ == "duck"


def test_no_access_to_class_property(db: SQLAlchemy) -> None:
    """Ensure the implementation doesn't access class properties or declared
    attrs while inspecting the unmapped model.
    """

    class class_property:
        def __init__(self, f: t.Callable[..., t.Any]) -> None:
            self.f = f

        def __get__(self, instance: t.Any, owner: t.Type[t.Any]) -> t.Any:
            return self.f(owner)

    class Duck(db.Model):
        id = sa.Column(sa.Integer, primary_key=True)

    class ns:
        is_duck = False
        floats = False

    class Witch(Duck):
        @sa.orm.declared_attr
        def is_duck(self) -> None:
            # declared attrs will be accessed during mapper configuration,
            # but make sure they're not accessed before that
            info = inspect.getouterframes(inspect.currentframe())[2]
            assert info[3] != "_should_set_tablename"
            ns.is_duck = True

        @class_property
        def floats(self) -> None:
            ns.floats = True

    assert ns.is_duck
    assert not ns.floats


def test_metadata_has_table(db: SQLAlchemy) -> None:
    user = db.Table("user", sa.Column("id", sa.Integer, primary_key=True))

    class User(db.Model):
        pass

    assert User.__table__ is user


def test_correct_error_for_no_primary_key(db: SQLAlchemy) -> None:
    with pytest.raises(sa.exc.ArgumentError) as info:

        class User(db.Model):
            pass

    assert "could not assemble any primary key" in str(info.value)


def test_single_has_parent_table(db: SQLAlchemy) -> None:
    class Duck(db.Model):
        id = sa.Column(sa.Integer, primary_key=True)

    class Call(Duck):
        pass

    assert Call.__table__ is Duck.__table__
    assert "__table__" not in Call.__dict__



from __future__ import annotations

import typing as t
from typing import Optional

import pytest
from quart import Quart
from werkzeug.exceptions import NotFound

from quart_sqlalchemy import SQLAlchemy
from quart_sqlalchemy.pagination import Pagination


class RangePagination(Pagination):
    def __init__(
        self, total: Optional[int] = 150, page: int = 1, per_page: int = 10
    ) -> None:
        if total is None:
            self._data = range(150)
        else:
            self._data = range(total)

        super().__init__(total=total, page=page, per_page=per_page)

        if total is None:
            self.total = None

    def _query_items(self) -> list[t.Any]:
        first = self._query_offset
        last = first + self.per_page + 1
        return list(self._data[first:last])

    def _query_count(self) -> int:
        return len(self._data)


def test_first_page() -> None:
    p = RangePagination()
    assert p.page == 1
    assert p.per_page == 10
    assert p.total == 150
    assert p.pages == 15
    assert not p.has_prev
    assert p.prev_num is None
    assert p.has_next
    assert p.next_num == 2


def test_last_page() -> None:
    p = RangePagination(page=15)
    assert p.page == 15
    assert p.has_prev
    assert p.prev_num == 14
    assert not p.has_next
    assert p.next_num is None


def test_item_numbers_first_page() -> None:
    p = RangePagination()
    p.items = list(range(10))
    assert p.first == 1
    assert p.last == 10


def test_item_numbers_last_page() -> None:
    p = RangePagination(page=15)
    p.items = list(range(5))
    assert p.first == 141
    assert p.last == 145


def test_item_numbers_0() -> None:
    p = RangePagination(total=0)
    assert p.first == 0
    assert p.last == 0


@pytest.mark.parametrize("total", [0, None])
def test_0_pages(total: int | None) -> None:
    p = RangePagination(total=total)
    assert p.pages == 0
    assert not p.has_prev
    assert not p.has_next


@pytest.mark.parametrize(
    ("page", "expect"),
    [
        (1, [1, 2, 3, 4, 5, None, 14, 15]),
        (2, [1, 2, 3, 4, 5, 6, None, 14, 15]),
        (3, [1, 2, 3, 4, 5, 6, 7, None, 14, 15]),
        (4, [1, 2, 3, 4, 5, 6, 7, 8, None, 14, 15]),
        (5, [1, 2, 3, 4, 5, 6, 7, 8, 9, None, 14, 15]),
        (6, [1, 2, None, 4, 5, 6, 7, 8, 9, 10, None, 14, 15]),
        (7, [1, 2, None, 5, 6, 7, 8, 9, 10, 11, None, 14, 15]),
        (8, [1, 2, None, 6, 7, 8, 9, 10, 11, 12, None, 14, 15]),
        (9, [1, 2, None, 7, 8, 9, 10, 11, 12, 13, 14, 15]),
        (10, [1, 2, None, 8, 9, 10, 11, 12, 13, 14, 15]),
        (11, [1, 2, None, 9, 10, 11, 12, 13, 14, 15]),
        (12, [1, 2, None, 10, 11, 12, 13, 14, 15]),
        (13, [1, 2, None, 11, 12, 13, 14, 15]),
        (14, [1, 2, None, 12, 13, 14, 15]),
        (15, [1, 2, None, 13, 14, 15]),
    ],
)
def test_iter_pages(page: int, expect: list[int | None]) -> None:
    p = RangePagination(page=page)
    assert list(p.iter_pages()) == expect


def test_iter_0_pages() -> None:
    p = RangePagination(total=0)
    assert list(p.iter_pages()) == []


@pytest.mark.parametrize("page", [1, 2, 3, 4])
def test_iter_pages_short(page: int) -> None:
    p = RangePagination(page=page, total=40)
    assert list(p.iter_pages()) == [1, 2, 3, 4]


class _PaginateCallable:
    def __init__(self, app: Quart, db: SQLAlchemy, Todo: t.Any) -> None:
        self.app = app
        self.db = db
        self.Todo = Todo

    async def __call__(
        self,
        page: Optional[int] = 1,
        per_page: Optional[int] = 20,
        max_per_page: Optional[int] = None,
        error_out: bool = True,
        count: bool = True,
    ) -> Pagination:
        qs = {"page": page, "per_page": per_page}
        async with self.app.test_request_context("/", query_string=qs):
            return self.db.paginate(
                self.db.select(self.Todo),
                max_per_page=max_per_page,
                error_out=error_out,
                count=count,
            )


@pytest.fixture
async def paginate(app: Quart, db: SQLAlchemy, Todo: t.Any) -> _PaginateCallable:
    async with app.app_context():
        for i in range(1, 101):
            db.session.add(Todo(title=f"task {i}"))

        db.session.commit()

    return _PaginateCallable(app, db, Todo)


async def test_paginate(paginate: _PaginateCallable) -> None:
    p = await paginate()
    assert p.page == 1
    assert p.per_page == 20
    assert len(p.items) == 20
    assert p.total == 100
    assert p.pages == 5


async def test_paginate_qs(paginate: _PaginateCallable) -> None:
    p = await paginate(page=2, per_page=10)
    assert p.page == 2
    assert p.per_page == 10


async def test_paginate_max(paginate: _PaginateCallable) -> None:
    p = await paginate(per_page=100, max_per_page=50)
    assert p.per_page == 50


async def test_no_count(paginate: _PaginateCallable) -> None:
    p = await paginate(count=False)
    assert p.total is None


@pytest.mark.parametrize(
    ("page", "per_page"), [("abc", None), (None, "abc"), (0, None), (None, -1)]
)
async def test_error_out(paginate: _PaginateCallable, page: t.Any, per_page: t.Any) -> None:
    with pytest.raises(NotFound):
        await paginate(page=page, per_page=per_page)


async def test_no_items_404(app: Quart, db: SQLAlchemy, Todo: t.Any) -> None:
    async with app.app_context():
        p = db.paginate(db.select(Todo))
        assert len(p.items) == 0

        with pytest.raises(NotFound):
            db.paginate(db.select(Todo), page=2)



from dataclasses import dataclass
from dataclasses import field
from functools import reduce
from typing import Mapping
from typing import Sequence
from typing import Union

import sqlalchemy as sa
from sqlalchemy import Table
from sqlalchemy.sql.elements import BinaryExpression
from sqlalchemy.sql.elements import ClauseElement


Filters = Sequence[BinaryExpression]
Columns = Sequence[Union[sa.Column, str]]

# ClauseElement is a weird name for sql-alchemy query, that's why SqlAlchemyQuery was defined
# Union used because SqlAlchemyQuery = ClauseElement raises mypy error:
# https://github.com/python/mypy/issues/7866
SqlAlchemyQuery = Union[ClauseElement]


@dataclass
class SelectQuery:
    """SQL SELECT query with customizable filters, orders, columns"""

    table: Table
    filters: Filters = field(default_factory=list)
    orders: Columns = field(default_factory=list)
    select_columns: Columns = field(default_factory=list)

    def __call__(self) -> SqlAlchemyQuery:
        """Create SQL SELECT query"""
        select_columns = self.select_columns if self.select_columns else [self.table]
        query = sa.select(select_columns)
        query = self.apply_filters(query, self.filters)
        query = self.apply_orders(query, self.orders)
        return query

    @staticmethod
    def apply_filters(query: SqlAlchemyQuery, filters: Filters) -> ClauseElement:
        """Append WHERE clause to query"""
        return reduce(lambda query_, filter_: query_.where(filter_), filters, query)

    @staticmethod
    def apply_orders(query: SqlAlchemyQuery, orders: Columns) -> ClauseElement:
        """Append ORDER BY clause to query"""
        return reduce(lambda query_, order_by: query_.order_by(order_by), orders, query)


@dataclass
class InsertQuery:
    """SQL INSERT query with customizable insert values and returning columns"""

    table: Table
    insert_values: Mapping
    returning_columns: Columns = field(default_factory=list)

    def __call__(self) -> SqlAlchemyQuery:
        """Create SQL INSERT query"""
        query = self.table.insert().values(self.insert_values)
        if self.returning_columns:
            query = query.returning(*self.returning_columns)
        return query


@dataclass
class InsertManyQuery:
    """Same as InsertQuery, but insert multiple values"""

    table: Table
    insert_values: Sequence[Mapping]
    returning_columns: Columns = field(default_factory=list)

    def __call__(self) -> SqlAlchemyQuery:
        """Create SQL INSERT query"""
        query = self.table.insert().values(self.insert_values)
        if self.returning_columns:
            query = query.returning(*self.returning_columns)
        return query


@dataclass
class UpdateQuery:
    """SQL UPDATE query with customizable update values and filters"""

    table: Table
    update_values: Mapping
    filters: Filters

    def __call__(self) -> SqlAlchemyQuery:
        """Create SQL UPDATE query"""
        return SelectQuery.apply_filters(
            self.table.update().values(self.update_values), self.filters
        )

    @classmethod
    def by_id(
        cls, id_: int, table: Table, update_values: Mapping, extra_filters: Filters = None
    ) -> "UpdateQuery":
        """Create update query with id filter"""
        extra_filters = extra_filters or []
        return UpdateQuery(table, update_values, [table.c.id == id_, *extra_filters])


@dataclass
class DeleteQuery:
    """SQL DELETE query with customizable filters"""

    table: Table
    filters: Union[Filters, Sequence[None]]

    def __call__(self) -> SqlAlchemyQuery:
        """
        Create SQL DELETE query
        :raise ValueError if {filters} is empty. To delete all table rows pass {filters} = [None]
        """
        if not len(self.filters):
            raise ValueError(
                """No filters set, are you sure you want to delete all table rows?
            If so call the method with None:
            repo.delete(None)"""
            )

        # None passed => delete all table rows
        filters: Filters
        if self.filters[0] is None:
            filters = tuple()
        else:
            filters = self.filters

        query = self.table.delete()
        query = SelectQuery.apply_filters(query, filters)
        return query



from __future__ import annotations

from typing import Any

from quart import Quart

from quart_sqlalchemy import SQLAlchemy
from quart_sqlalchemy.cli import add_models_to_shell


async def test_shell_context(app: Quart, db: SQLAlchemy, Todo: Any) -> None:
    async with app.app_context():
        context = add_models_to_shell()
        assert context["db"] is db
        assert context["Todo"] is Todo



import asyncio

from quart import Quart
from sqlalchemy.future import select
from sqlalchemy.orm import selectinload
from sqlalchemy.sql import func

from quart_sqlalchemy import SQLAlchemy


def create_app():
    app = Quart(__name__)
    app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite+aiosqlite:///:memory:"
    app.config["SQLALCHEMY_ENGINE_ASYNC"] = True
    return app


app = create_app()
db = SQLAlchemy(app)


class A(db.Model):
    __tablename__ = "a"

    id = db.Column(db.Integer, primary_key=True)
    data = db.Column(db.String)
    create_date = db.Column(db.DateTime, server_default=func.now())
    bs = db.relationship("B")


class B(db.Model):
    __tablename__ = "b"
    id = db.Column(db.Integer, primary_key=True)
    a_id = db.Column(db.ForeignKey("a.id"))
    data = db.Column(db.String)


async def async_main():
    import pdb

    pdb.set_trace()
    async with app.app_context():
        async with db.engine.begin() as conn:
            await conn.run_sync(db.metadata.drop_all)
            await conn.run_sync(db.metadata.create_all)

        # async with db.engine.begin() as conn:
        async with db.session() as session:
            async with session.begin():
                session.add(A(bs=[B(), B()], data="a1"))
                # session.add_all(
                #     [
                #         A(bs=[B(), B()], data="a1"),
                #         A(bs=[B()], data="a2"),
                #         A(bs=[B(), B()], data="a3"),
                #     ]
                # )
                await session.commit()
                # await session.commit()

    # async with app.app_context():
    #         stmt = select(A).options(selectinload(A.bs))

    #         result = await session.execute(stmt)

    #         for a1 in result.scalars():
    #             print(a1)
    #             print(f"created at: {a1.create_date}")
    #             for b1 in a1.bs:
    #                 print(b1)

    # await db.engine.dispose()


asyncio.run(async_main())



from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from typing import Any
from typing import AsyncIterator
from typing import cast
from typing import Dict
from typing import Generic
from typing import List
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import Type
from typing import TypeVar

import sqlalchemy as sa
import typing_inspect
from pydantic import BaseModel
from sql_repo.repo.repk_old.impl import Columns
from sql_repo.repo.repk_old.impl import DeleteQuery
from sql_repo.repo.repk_old.impl import Filters
from sql_repo.repo.repk_old.impl import InsertManyQuery
from sql_repo.repo.repk_old.impl import InsertQuery
from sql_repo.repo.repk_old.impl import SelectQuery
from sql_repo.repo.repk_old.impl import SqlAlchemyQuery
from sql_repo.repo.repk_old.impl import UpdateQuery
from sql_repo.utils import aiter_to_list
from sql_repo.utils import is_field_equal_to_default
from sql_repo.utils import mixed_zip
from sql_repo.utils import model_to_primitive
from sqlalchemy import Table
from sqlalchemy.sql.elements import BinaryExpression


Created = bool


class IdModel(BaseModel):
    """Pydantic model with optional id field"""

    id: Optional[int]


GenericIdModel = TypeVar("GenericIdModel", bound=IdModel)


class AsyncQueryExecutor:
    @abstractmethod
    async def fetch_one(self, query: SqlAlchemyQuery, **sa_params: Any) -> Optional[Mapping]:
        """Execute SELECT query and return first result row"""

    @abstractmethod
    async def fetch_all(self, query: SqlAlchemyQuery, **sa_params: Any) -> AsyncIterator[Mapping]:
        """Execute SELECT query and return all result rows"""

    @abstractmethod
    async def fetch_val(self, query: SqlAlchemyQuery, **sa_params: Any) -> Any:
        """Execute SELECT query and return first column of first result row"""

    @abstractmethod
    async def insert(self, query: SqlAlchemyQuery, **sa_params: Any) -> Mapping:
        """Execute INSERT query and return returning columns"""

    @abstractmethod
    async def insert_many(self, query: SqlAlchemyQuery, **sa_params: Any) -> AsyncIterator[Mapping]:
        """Execute INSERT query and return list of returning columns"""

    @abstractmethod
    async def update(self, query: SqlAlchemyQuery, **sa_params: Any) -> None:
        """Execute UPDATE query"""

    @abstractmethod
    async def delete(self, query: SqlAlchemyQuery, **sa_params: Any) -> None:
        """Execute DELETE query"""

    @abstractmethod
    def execute_in_transaction(self) -> Any:
        """Execute queries in transaction"""


class AsyncBaseRepo(Generic[GenericIdModel], ABC):
    """
    Execute sql-queries, convert sql-row-dicts to/from pydantic models in async way
    """

    # =============
    # CONFIGURATION
    # =============

    @property
    @abstractmethod
    def table(self) -> Table:
        """SQLAlchemy table definition"""

    @property
    def ignore_default(self) -> Sequence[str]:
        """
        Columns will be inserted only if their values are not equal to default values of
        corresponding models' fields
        These columns will be set after insert
        """
        return []

    def serialize(self, entity: GenericIdModel) -> Dict:
        """Convert pydantic model to dict"""
        return model_to_primitive(entity, without_id=True)

    def deserialize(self, **kwargs: Any) -> GenericIdModel:
        """Create pydantic model from kwargs"""
        entity_type = self._get_generic_type()
        return entity_type(**kwargs)

    @property
    @abstractmethod
    def query_executor(self) -> AsyncQueryExecutor:
        """repka.repositories.base.AsyncQueryExecutor instance"""

    # ==============
    # SELECT METHODS
    # ==============

    async def first(
        self, *filters: BinaryExpression, orders: Columns = None
    ) -> Optional[GenericIdModel]:
        """Get first entity from DB matching filters and orders"""
        query = SelectQuery(self.table, filters, orders or [])()
        row = await self.query_executor.fetch_one(query)
        return self.deserialize(**row) if row else None

    async def get_by_id(self, entity_id: int) -> Optional[GenericIdModel]:
        """Get entity from DB with id = {entity_id}"""
        return await self.first(self.table.c.id == entity_id)

    async def get_or_create(
        self, filters: Filters = None, defaults: Dict = None
    ) -> Tuple[GenericIdModel, Created]:
        """Get first entity from DB  matching filters or create it"""
        entity = await self.first(*(filters or []))
        if entity:
            return entity, False

        entity = self.deserialize(**(defaults or {}))
        entity = await self.insert(entity)
        return entity, True

    async def get_all(
        self, filters: Filters = None, orders: Columns = None
    ) -> List[GenericIdModel]:
        """Get all entities from DB matching filters and orders"""
        return await aiter_to_list(await self.get_all_aiter(filters, orders))

    async def get_all_aiter(
        self, filters: Filters = None, orders: Columns = None
    ) -> AsyncIterator[GenericIdModel]:
        """Get all entities from DB matching filters and orders as an async iterator"""
        query = SelectQuery(self.table, filters or [], orders or [])()
        rows = await self.query_executor.fetch_all(query)
        return self._rows_to_entities(rows)

    async def get_by_ids(self, entity_ids: Sequence[int]) -> List[GenericIdModel]:
        """Get all entities from DB with id from {entity_ids}"""
        return await aiter_to_list(await self.get_by_ids_aiter(entity_ids))

    async def get_by_ids_aiter(self, entity_ids: Sequence[int]) -> AsyncIterator[GenericIdModel]:
        """Get all entities from DB with id from {entity_ids} as an async iterator"""
        return await self.get_all_aiter(filters=[self.table.c.id.in_(entity_ids)])

    async def get_all_ids(
        self, filters: Sequence[BinaryExpression] = None, orders: Columns = None
    ) -> Sequence[int]:
        """Same as get_all() but returns only ids."""
        query = SelectQuery(
            self.table, filters or [], orders or [], select_columns=[self.table.c.id]
        )()
        rows = await self.query_executor.fetch_all(query)
        return [row["id"] async for row in rows]

    async def exists(self, *filters: BinaryExpression) -> bool:
        """Check entity matching filters exists in DB"""
        query = SelectQuery(self.table, filters, select_columns=[sa.func.count(self.table.c.id)])()
        result = await self.query_executor.fetch_val(query)
        return bool(result)

    # ==============
    # INSERT METHODS
    # ==============

    async def insert(self, entity: GenericIdModel) -> GenericIdModel:
        """Insert entity to DB"""
        return await InsertImpl(self).insert(entity)

    async def insert_many(self, entities: List[GenericIdModel]) -> List[GenericIdModel]:
        """Insert multiple entities to DB"""
        return await InsertManyImpl(self).insert_many(entities)

    async def insert_many_aiter(
        self, entities: List[GenericIdModel]
    ) -> AsyncIterator[GenericIdModel]:
        """Insert multiple entities to DB. Returns an async iterable with inserted entities"""
        return await InsertManyImpl(self).insert_many_aiter(entities)

    # ==============
    # UPDATE METHODS
    # ==============

    async def update(self, entity: GenericIdModel) -> GenericIdModel:
        """Update entity in DB"""
        assert entity.id
        update_values = self.serialize(entity)
        query = UpdateQuery.by_id(entity.id, self.table, update_values)()
        await self.query_executor.update(query)
        return entity

    async def update_partial(self, entity: GenericIdModel, **updated_values: Any) -> GenericIdModel:
        """Update particular entity fields in DB"""
        assert entity.id

        for field, value in updated_values.items():
            setattr(entity, field, value)

        serialized_entity = self.serialize(entity)
        serialized_values = {key: serialized_entity[key] for key in updated_values.keys()}

        query = UpdateQuery.by_id(entity.id, self.table, serialized_values)()
        await self.query_executor.update(query)

        return entity

    async def update_many(self, entities: List[GenericIdModel]) -> List[GenericIdModel]:
        """
        Update multiple entities in DB
        No way to update many in single query:
        https://github.com/aio-libs/aiopg/issues/546
        So update entities sequentially in transaction.
        """
        if not entities:
            return entities

        async with self.execute_in_transaction():
            entities = [await self.update(entity) for entity in entities]

        return entities

    async def update_values(self, values: dict, filters: Filters) -> None:
        """Update particular entity fields for entities matching filters (perform SQL UPDATE)"""
        query = UpdateQuery(self.table, values, filters)()
        await self.query_executor.update(query)

    async def update_or_insert_first_by_field(
        self, entity: GenericIdModel, field: str
    ) -> GenericIdModel:
        """Update one entity with field or add it to DB"""
        value = getattr(entity, field)
        entity_with_field = await self.first(self.table.c[field] == value)

        if entity_with_field:
            entity.id = entity_with_field.id
            entity = await self.update(entity)
        else:
            entity = await self.insert(entity)

        return entity

    async def update_or_insert_many_by_field(
        self, entities: Sequence[GenericIdModel], field: str
    ) -> Sequence[GenericIdModel]:
        """Update all entities with field and add entities without it to DB"""
        values = [getattr(e, field) for e in entities]
        entities_with_field = await self.get_all(filters=[self.table.c[field].in_(values)])
        field_entities = {getattr(e, field): e for e in entities_with_field}

        async with self.execute_in_transaction():
            entities_to_insert, entities_to_update = [], []
            for e in entities:
                if getattr(e, field) in field_entities:
                    entities_to_update.append(e)
                else:
                    entities_to_insert.append(e)

            entities_to_insert = await self.insert_many(entities_to_insert)

            for e in entities_to_update:
                e.id = field_entities[getattr(e, field)].id
            entities_to_update = await self.update_many(entities_to_update)

        return [*entities_to_insert, *entities_to_update]

    # ==============
    # DELETE METHODS
    # ==============

    async def delete(self, *filters: Optional[BinaryExpression]) -> None:
        """Delete entities matching filters from DB"""
        query = DeleteQuery(self.table, filters)()
        await self.query_executor.delete(query)

    async def delete_by_id(self, entity_id: int) -> None:
        """Delete entity by id from DB"""
        return await self.delete(self.table.c.id == entity_id)

    async def delete_by_ids(self, entity_ids: Sequence[int]) -> None:
        """Delete multiple entities from DB with id in {entity_ids}"""
        return await self.delete(self.table.c.id.in_(entity_ids))

    # ==============
    # OTHER METHODS
    # ==============

    def execute_in_transaction(self) -> Any:
        """
        Execute queries in transaction
        Usage:
        async with repo.execute_in_transaction():
            repo.delete(...)
            repo.insert(...)
            ...
        """
        return self.query_executor.execute_in_transaction()

    # ==============
    # PROTECTED & PRIVATE METHODS
    # ==============

    def _get_generic_type(self) -> Type[GenericIdModel]:
        """
        Get generic type of inherited BaseRepository:
        >>> class TransactionRepo(AiopgRepository[Transaction]):
        ...     table = transactions_table
        ... # doctest: +SKIP
        >>> assert TransactionRepo().__get_generic_type() is Transaction # doctest: +SKIP
        """
        return cast(
            Type[GenericIdModel],
            typing_inspect.get_args(typing_inspect.get_generic_bases(self)[-1])[0],
        )

    async def _rows_to_entities(
        self, rows: AsyncIterator[Mapping]
    ) -> AsyncIterator[GenericIdModel]:
        """
        Converts an async iterator of DB rows to an async iterator of GenericIdModel
        """
        async for row in rows:
            yield cast(GenericIdModel, self.deserialize(**row))


@dataclass
class InsertImpl:
    """Entity DB insertion implementation"""

    repo: AsyncBaseRepo

    async def insert(self, entity: GenericIdModel) -> GenericIdModel:
        """Perform entity insertion"""
        query = InsertQuery(
            self.repo.table, self._serialize_for_insertion(entity), self.insert_returning_columns
        )()

        row = await self.repo.query_executor.insert(query)

        return self._set_ignored_fields(entity, row)

    def _serialize_for_insertion(self, entity: GenericIdModel) -> Dict[str, Any]:
        """
        Remove ignored fields from serialized entity
        Field should be removed here (not in .serialize) due to compatibility
        """
        return {
            key: value
            for key, value in self.repo.serialize(entity).items()
            if key not in self._get_ignored_fields(entity)
        }

    @property
    def insert_returning_columns(self) -> Columns:
        """All columns except columns from {AsyncBaseRepo.ignore_default}"""
        return (
            self.repo.table.c.id,
            *(getattr(self.repo.table.c, col) for col in self.repo.ignore_default),
        )

    def _set_ignored_fields(self, entity: GenericIdModel, row: Mapping) -> GenericIdModel:
        """Set returned from db values to entity"""
        entity.id = row["id"]
        for col in self._get_ignored_fields(entity):
            setattr(entity, col, row[col])
        return entity

    def _get_ignored_fields(self, entity: GenericIdModel) -> Set[str]:
        """Get entity fields which values equal to default value"""
        return {
            field for field in self.repo.ignore_default if is_field_equal_to_default(entity, field)
        }


class InsertManyImpl(InsertImpl):
    """Multiple entities DB insertion implementation"""

    async def insert_many(self, entities: List[GenericIdModel]) -> List[GenericIdModel]:
        """
        Inserts many entities with a single query.
        :raises ValueError if some entities' fields from self.ignore_default have default values
        while other fields have non-default values
        """
        return await aiter_to_list(await self.insert_many_aiter(entities))

    async def insert_many_aiter(
        self, entities: List[GenericIdModel]
    ) -> AsyncIterator[GenericIdModel]:
        if not entities:

            async def _empty_aiter() -> AsyncIterator[GenericIdModel]:
                return
                yield

            return _empty_aiter()

        self._check_server_defaults(entities)

        query = InsertManyQuery(
            self.repo.table,
            [self._serialize_for_insertion(entity) for entity in entities],
            self.insert_returning_columns,
        )()

        rows = await self.repo.query_executor.insert_many(query)

        return self._updated_entities_aiter(entities, rows)

    def _check_server_defaults(self, entities: Sequence[GenericIdModel]) -> None:
        """Check all entity values either equal to default values or not"""
        server_default_fields = [
            col.key
            for col in self.repo.table.c
            if col.server_default is not None and col.key in self.repo.ignore_default
        ]
        for server_default_field in server_default_fields:
            first = is_field_equal_to_default(next(iter(entities)), server_default_field)
            is_consistent = all(
                is_field_equal_to_default(entity, server_default_field) == first
                for entity in entities
            )
            if not is_consistent:
                raise ValueError(
                    "All fields from ignore default should either be equal to default values or not be "
                    f"equal. Got inconsistency with {server_default_field} field"
                )

    async def _updated_entities_aiter(
        self, entities: List[GenericIdModel], rows: AsyncIterator[Mapping]
    ) -> AsyncIterator[GenericIdModel]:
        """
        Returns an async iterator which yielding entities with fields updated by passed rows
        """
        async for entity, row in mixed_zip(entities, rows):  # type: ignore
            yield self._set_ignored_fields(entity, row)



import base62
from reedsolo import RSCodec
from speck import SpeckCipher


key_96bit = 0x9E4DC9158C2D33AAD4A9BEE3C3F2F134DBAD

speck = SpeckCipher(key_96bit, key_size=96, block_size=48)
reed_solomon_codec = RSCodec(10)


def encode(value: int, ecc: bool = True) -> str:
    ciphertext = speck.encrypt(value)
    cipherbytes = ciphertext.to_bytes(6, "big")
    if ecc:
        cipherbytes = bytes(reed_solomon_codec.encode(cipherbytes))
    return base62.encodebytes(cipherbytes)


def decode(encoded: str, ecc: bool = True) -> int:
    cipherbytes = base62.decodebytes(encoded)
    if ecc:
        cipherbytes, *_ = reed_solomon_codec.decode(bytearray(cipherbytes))
    ciphertext = int.from_bytes(cipherbytes, "big")
    return speck.decrypt(ciphertext)


class OID:
    def __init__(self, value):
        if isinstance(value, OID):
            value = value._value
        if isinstance(value, str):
            value = decode(value)

        self._value = value

    def decode(self) -> int:
        return decode(self._value)

    def encode(self) -> str:
        return encode(self._value)


# model_id = 99999999999

# # 235963431219045
# ciphertext = speck.encrypt(model_id)

# #  b'\xd6\x9b\x84\x7foe'
# cipherbytes = ciphertext.to_bytes(6, "big")

# # # '150GeEMpZ'
# # encoded = base62.encodebytes(cipherbytes)

# # bytearray(b'\xd6\x9b\x84\x7foe\xefV\xbe\x02\x9e\x1b\x9cJ[\xb1')
# ecc_encoded = rsc.encode(cipherbytes)

# # '6WxQdmGgs2xKd1OHtlYdeb'
# base62_encoded = base62.encodebytes(bytes(ecc_encoded))


# # lets decode and decrypt back to original id

# # 1. With ECC

# # b'\xd6\x9b\x84\x7foe\xefV\xbe\x02\x9e\x1b\x9cJ[\xb1'
# decoded_ecc = base62.decodebytes(base62_encoded)

# # bytearray(b'\xd6\x9b\x84\x7foe')
# cipherbytes, *_ = rsc.decode(bytearray(decoded_ecc))

# # 235963431219045
# ciphertext = int.from_bytes(cipherbytes, "big")

# # 99999999999
# recovered_model_id = speck.decrypt(ciphertext)



from datetime import datetime
from datetime import timezone
import asyncio

from quart import flash
from quart import Quart
from quart import redirect
from quart import render_template
from quart import request
from quart import url_for
from quart_sqlalchemy import SQLAlchemy

app = Quart(__name__)
app.secret_key = "Achee6phIexoh8dagiQuew0ephuga4Ih"
app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite:///todo.sqlite"
app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False
db = SQLAlchemy(app, engine_options=dict(check_same_thread=False))


def now_utc():
    return datetime.now(timezone.utc)


class Todo(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    title = db.Column(db.String, nullable=False)
    text = db.Column(db.String, nullable=False)
    done = db.Column(db.Boolean, nullable=False, default=False)
    pub_date = db.Column(db.DateTime, nullable=False, default=now_utc)


@app.route("/")
async def show_all():
    select = db.select(Todo).order_by(Todo.pub_date.desc())
    todos = db.session.execute(select).all()
    # import pdb; pdb.set_trace()
    return dict(status='ok', todos=len(todos))
    # return await render_template("show_all.html", todos=todos)


@app.route("/new", methods=["GET", "POST"])
async def new():
    if request.method == "POST":
        form = await request.form
        if not form["title"]:
            await flash("Title is required", "error")
        elif not form["text"]:
            await flash("Text is required", "error")
        else:
            todo = Todo(title=form["title"], text=form["text"])
            db.session.add(todo)
            db.session.commit()
            await flash("Todo item was successfully created")
            return redirect(url_for("show_all"))

    return await render_template("new.html")


@app.route("/update", methods=["POST"])
async def update_done():
    form = await request.form

    for todo in db.session.execute(db.select(Todo)).scalars():
        todo.done = f"done.{todo.id}" in form

    await flash("Updated status")
    db.session.commit()
    return redirect(url_for("show_all"))


async def setup_db():
    async with app.app_context():
        db.create_all()


if __name__ == "__main__":
    asyncio.run(setup_db())